# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Callable, Generic, TypeVar, Union, overload

import torch
import torch.distributed as dist

T = TypeVar("T", bound=Callable[..., Any])

try:
    import nemo_run as run

    Config = run.Config
    Partial = run.Partial
except ImportError:
    logging.warning(
        "Trying to use Config or Partial, but NeMo-Run is not installed. Please install NeMo-Run before proceeding."
    )

    _T = TypeVar("_T")

    class Config(Generic[_T]):
        """ """

        pass

    class Partial(Generic[_T]):
        """ """

        pass


SAFE_REPOS = [
    "nvidia",
    "Qwen",
    "deepseek-ai",
    "meta-llama",
    "google",
    "openai",
    "mistralai",
    "moonshotai",
    "llava-hf",
    "gpt2",
    "baichuan-inc",
]


def task(*args: Any, **kwargs: Any) -> Callable[[T], T]:
    """ """

    try:
        import nemo_run as run

        return run.task(*args, **kwargs)
    except (ImportError, AttributeError):
        # Return a no-op function
        def noop_decorator(func: T) -> T:
            return func

        return noop_decorator


@overload
def factory() -> Callable[[T], T]: ...


@overload
def factory(*args: Any, **kwargs: Any) -> Callable[[T], T]: ...


def factory(*args: Any, **kwargs: Any) -> Union[Callable[[T], T], T]:
    """ """

    try:
        import nemo_run as run

        if not args:
            return run.factory(**kwargs)
        else:
            # Used as @factory(*args, **kwargs)
            return run.factory(*args, **kwargs)
    except (ImportError, AttributeError):
        # Return a no-op function
        def noop_decorator(func: T) -> T:
            return func

        if not args and not kwargs:
            return noop_decorator
        else:
            return noop_decorator


def torch_dtype_from_precision(precision: Union[int, str]) -> torch.dtype:
    """Mapping from PTL precision types to corresponding PyTorch parameter datatype."""

    if precision in ('bf16', 'bf16-mixed'):
        return torch.bfloat16
    elif precision in (16, '16', '16-mixed'):
        return torch.float16
    elif precision in (32, '32', '32-true'):
        return torch.float32
    else:
        raise ValueError(f"Could not parse the precision of `{precision}` to a valid torch.dtype")


def barrier():
    """Waits for all processes."""
    if dist.is_initialized():
        dist.barrier()


def is_safe_repo(hf_path: str, trust_remote_code: bool | None) -> bool:
    """
    Decide whether remote code execution should be enabled for a Hugging Face
    model or dataset repository.
    This function follows three rules:
        1. If `trust_remote_code` is explicitly provided (True/False), its value
            takes precedence.
        2. If `trust_remote_code` is None, the function checks whether the repo
            belongs to a predefined list of trusted repositories (`SAFE_REPOS`).
        3. Otherwise, remote code execution is disabled.
    Args:
        hf_path (str):
            The Hugging Face repository identifier (e.g., "org/model_name").
        trust_remote_code (bool | None):
            If True, always allow remote code execution.
            If False, always disable it.
            If None, fall back to internal safety rules and trusted repo list.
    Returns:
        bool: Whether remote code execution should be enabled.
    """
    if trust_remote_code is not None:
        if trust_remote_code is False:
            logging.warning(
                "`trust_remote_code=False`. Remote code may not be executed. "
                "Set `trust_remote_code=True` only if you fully trust the Hugging Face repository."
            )
        return trust_remote_code

    hf_repo = hf_path.split("/")[0]
    if hf_repo in SAFE_REPOS:
        return True

    return False



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch.testing

from nemo.collections.common.prompts.canary import PromptFormatter
from nemo.collections.common.prompts.formatter import Modality


class _DummyPromptFormatter(PromptFormatter):
    NAME = "_dummy_test_formatter"
    TEMPLATE = {
        "user": {"template": "<s>|text|</s>", "slots": {"text": Modality.Text}},
        "assistant": {"template": "|text|</s>", "slots": {"text": Modality.Text}},
    }
    OUTPUT_ROLE = "assistant"


def test_prompt_formatter_empty_dialog_exception(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    with pytest.raises(AssertionError):
        formatter.encode_dialog([])


def test_prompt_formatter_inference(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog([{"role": "user", "slots": {"text": "hi"}}])
    recovered = bpe_tokenizer.ids_to_text(ans["input_ids"])
    assert recovered == "<s>hi</s>"


def test_prompt_formatter_inference_using_content(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog([{"role": "user", "slots": {"text": "hi"}}])
    ans2 = formatter.encode_dialog([{"role": "user", "content": "hi"}])
    assert ans.keys() == ans2.keys()
    for k in ans:
        torch.testing.assert_close(ans[k], ans2[k])


def test_prompt_formatter_training(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog(
        [
            {"role": "user", "slots": {"text": "hi"}},
            {"role": "assistant", "slots": {"text": "hello"}},
        ]
    )
    recovered = bpe_tokenizer.ids_to_text(ans["input_ids"])
    assert recovered == "<s>hi</s> hello</s>", recovered


def test_prompt_formatter_training_using_content(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog(
        [
            {"role": "user", "slots": {"text": "hi"}},
            {"role": "assistant", "slots": {"text": "hello"}},
        ]
    )
    ans2 = formatter.encode_dialog(
        [
            {"role": "user", "content": "hi"},
            {"role": "assistant", "content": "hello"},
        ]
    )
    assert ans.keys() == ans2.keys()
    for k in ans:
        torch.testing.assert_close(ans[k], ans2[k])


def test_prompt_formatter_missing_role(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    with pytest.raises(AssertionError, match="A turn must have have a 'role' key"):
        formatter.encode_dialog([{"slots": {"text": "hi"}}])


def test_prompt_formatter_missing_slots(bpe_tokenizer):
    formatter = _DummyPromptFormatter(bpe_tokenizer)
    with pytest.raises(
        AssertionError, match="A turn for role user must have have a non-empty value under 'slots' key"
    ):
        formatter.encode_dialog([{"role": "user"}])
    with pytest.raises(
        AssertionError, match="A turn for role user must have have a non-empty value under 'slots' key"
    ):
        formatter.encode_dialog([{"role": "user", "slots": {}}])


def test_prompt_formatter_aggregate_tokenizer(canary_tokenizer):
    # Note the 'canary_tokenizer' arg which is an aggregate tokenizer fixture.
    formatter = _DummyPromptFormatter(canary_tokenizer)
    ans = formatter.encode_dialog(
        [
            {
                "role": "user",
                "slots": {
                    "text": "hi",
                    "prompt_language": "en",
                },
            }
        ]
    )
    recovered = canary_tokenizer.ids_to_text(ans["input_ids"])
    assert recovered == " <s>hi</s>"


def test_prompt_formatter_aggregate_tokenizer_missing_prompt_language(canary_tokenizer):
    # Note the 'canary_tokenizer' arg which is an aggregate tokenizer fixture.
    formatter = _DummyPromptFormatter(canary_tokenizer)

    with pytest.raises(AssertionError, match="Missing key 'prompt_language' in slot_values"):
        formatter.encode_dialog([{"role": "user", "slots": {"text": "hi"}}])


class _DummyPreamblePromptFormatter(PromptFormatter):
    NAME = "_dummy_preamble_test_formatter"
    TEMPLATE = {
        "preamble": {"template": "TEST"},
        "user": {"template": "<s>|text|</s>", "slots": {"text": Modality.Text}},
        "assistant": {"template": "|text|</s>", "slots": {"text": Modality.Text}},
    }
    OUTPUT_ROLE = "assistant"


def test_prompt_formatter_preamble_inference(bpe_tokenizer):
    formatter = _DummyPreamblePromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog([{"role": "user", "slots": {"text": "hi"}}])
    recovered = bpe_tokenizer.ids_to_text(ans["input_ids"])
    assert recovered == "TEST <s>hi</s>", recovered


def test_prompt_formatter_premble_training(bpe_tokenizer):
    formatter = _DummyPreamblePromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog(
        [
            {"role": "user", "slots": {"text": "hi"}},
            {"role": "assistant", "slots": {"text": "hello"}},
        ]
    )
    recovered = bpe_tokenizer.ids_to_text(ans["input_ids"])
    assert recovered == "TEST <s>hi</s> hello</s>"


def test_prompt_formatter_explicit_preamble(bpe_tokenizer):
    formatter = _DummyPreamblePromptFormatter(bpe_tokenizer)
    ans = formatter.encode_dialog([{"role": "preamble"}, {"role": "user", "slots": {"text": "hi"}}])
    recovered = bpe_tokenizer.ids_to_text(ans["input_ids"])
    assert recovered == "TEST <s>hi</s>"


def test_prompt_formatter_wrong_preamble_excpetions(bpe_tokenizer):
    formatter = _DummyPreamblePromptFormatter(bpe_tokenizer)
    with pytest.raises(AssertionError):
        # Error: 2 preambles
        formatter.encode_dialog(
            [
                {"role": "preamble"},
                {"role": "preamble"},
                {"role": "user", "slots": {"text": "hi"}},
            ]
        )
    with pytest.raises(AssertionError):
        # Error: preamble not at the beginning
        formatter.encode_dialog(
            [
                {"role": "user", "slots": {"text": "hi"}},
                {"role": "preamble"},
            ]
        )
    with pytest.raises(AssertionError):
        # Error: preamble with slots
        formatter.encode_dialog(
            [
                {"role": "user", "slots": {"text": "hi"}},
                {"role": "preamble", "slots": {"abc": "abc"}},
            ]
        )



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# MIT License
#
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Iterable, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat
from torch import Tensor

from nemo.collections.common.parts.utils import mask_sequence_tensor
from nemo.collections.tts.losses.audio_codec_loss import MaskedMSELoss
from nemo.collections.tts.modules.audio_codec_modules import (
    CodecActivation,
    Conv1dNorm,
    Conv2dNorm,
    ConvTranspose1dNorm,
    VectorQuantizerBase,
    get_down_sample_padding,
)
from nemo.collections.tts.parts.utils.distributed import broadcast_tensors
from nemo.core.classes.common import typecheck
from nemo.core.classes.module import NeuralModule
from nemo.core.neural_types.elements import AudioSignal, EncodedRepresentation, Index, LengthsType, LossType, VoidType
from nemo.core.neural_types.neural_type import NeuralType
from nemo.utils import logging


class SEANetResnetBlock(NeuralModule):
    def __init__(self, channels: int, activation: str = "elu"):
        super().__init__()
        self.pre_activation = CodecActivation(activation=activation, channels=channels)
        hidden_channels = channels // 2
        self.pre_conv = Conv1dNorm(in_channels=channels, out_channels=channels, kernel_size=1)
        self.res_conv1 = Conv1dNorm(in_channels=channels, out_channels=hidden_channels, kernel_size=3)
        self.post_activation = CodecActivation(activation=activation, channels=hidden_channels)
        self.res_conv2 = Conv1dNorm(in_channels=hidden_channels, out_channels=channels, kernel_size=1)

    @property
    def input_types(self):
        return {
            "inputs": NeuralType(('B', 'C', 'T_input'), VoidType()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        return {
            "out": NeuralType(('B', 'C', 'T_out'), VoidType()),
        }

    def remove_weight_norm(self):
        self.pre_conv.remove_weight_norm()
        self.res_conv1.remove_weight_norm()
        self.res_conv2.remove_weight_norm()

    @typecheck()
    def forward(self, inputs, input_len):
        res = self.pre_activation(inputs)
        res = self.res_conv1(inputs=res, input_len=input_len)
        res = self.post_activation(res)
        res = self.res_conv2(inputs=res, input_len=input_len)

        out = self.pre_conv(inputs=inputs, input_len=input_len) + res
        out = mask_sequence_tensor(out, input_len)
        return out


class SEANetRNN(NeuralModule):
    def __init__(self, dim: int, num_layers: int, rnn_type: str = "lstm", use_skip: bool = False):
        super().__init__()
        self.use_skip = use_skip
        if rnn_type == "lstm":
            self.rnn = torch.nn.LSTM(input_size=dim, hidden_size=dim, num_layers=num_layers)
        elif rnn_type == "gru":
            self.rnn = torch.nn.GRU(input_size=dim, hidden_size=dim, num_layers=num_layers)
        else:
            raise ValueError(f"Unknown RNN type {rnn_type}")

    @property
    def input_types(self):
        return {
            "inputs": NeuralType(('B', 'C', 'T'), VoidType()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        return {
            "out": NeuralType(('B', 'C', 'T'), VoidType()),
        }

    @typecheck()
    def forward(self, inputs, input_len):
        inputs = rearrange(inputs, "B C T -> B T C")

        packed_inputs = nn.utils.rnn.pack_padded_sequence(
            inputs, lengths=input_len.cpu(), batch_first=True, enforce_sorted=False
        )
        packed_out, _ = self.rnn(packed_inputs)
        out, _ = nn.utils.rnn.pad_packed_sequence(packed_out, batch_first=True)

        if self.use_skip:
            out = out + inputs

        out = rearrange(out, "B T C -> B C T")
        return out


class SEANetEncoder(NeuralModule):
    def __init__(
        self,
        down_sample_rates: Iterable[int] = (2, 4, 5, 8),
        base_channels: int = 32,
        in_kernel_size: int = 7,
        out_kernel_size: int = 7,
        encoded_dim: int = 128,
        activation: str = "elu",
        rnn_layers: int = 2,
        rnn_type: str = "lstm",
        rnn_skip: bool = True,
    ):
        assert in_kernel_size > 0
        assert out_kernel_size > 0

        super().__init__()

        self.down_sample_rates = down_sample_rates
        self.pre_conv = Conv1dNorm(in_channels=1, out_channels=base_channels, kernel_size=in_kernel_size)

        in_channels = base_channels
        self.res_blocks = nn.ModuleList([])
        self.down_sample_conv_layers = nn.ModuleList([])
        self.activations = nn.ModuleList([])
        for i, down_sample_rate in enumerate(self.down_sample_rates):
            res_block = SEANetResnetBlock(channels=in_channels)
            self.res_blocks.append(res_block)
            self.activations.append(CodecActivation(activation=activation, channels=in_channels))

            out_channels = 2 * in_channels
            kernel_size = 2 * down_sample_rate
            down_sample_conv = Conv1dNorm(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=down_sample_rate,
                padding=get_down_sample_padding(kernel_size, down_sample_rate),
            )
            in_channels = out_channels
            self.down_sample_conv_layers.append(down_sample_conv)

        self.post_activation = CodecActivation(activation=activation, channels=in_channels)
        self.rnn = SEANetRNN(dim=in_channels, num_layers=rnn_layers, rnn_type=rnn_type, use_skip=rnn_skip)
        self.post_conv = Conv1dNorm(in_channels=in_channels, out_channels=encoded_dim, kernel_size=out_kernel_size)

    @property
    def input_types(self):
        return {
            "audio": NeuralType(('B', 'T_audio'), AudioSignal()),
            "audio_len": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        return {
            "encoded": NeuralType(('B', 'D', 'T_encoded'), EncodedRepresentation()),
            "encoded_len": NeuralType(tuple('B'), LengthsType()),
        }

    def remove_weight_norm(self):
        self.pre_conv.remove_weight_norm()
        self.post_conv.remove_weight_norm()
        for res_block in self.res_blocks:
            res_block.remove_weight_norm()
        for down_sample_conv in self.down_sample_conv_layers:
            down_sample_conv.remove_weight_norm()

    @typecheck()
    def forward(self, audio, audio_len):
        encoded_len = audio_len
        audio = rearrange(audio, "B T -> B 1 T")
        # [B, C, T_audio]
        out = self.pre_conv(inputs=audio, input_len=encoded_len)
        for res_block, down_sample_conv, down_sample_rate, activation in zip(
            self.res_blocks, self.down_sample_conv_layers, self.down_sample_rates, self.activations
        ):
            # [B, C, T]
            out = res_block(inputs=out, input_len=encoded_len)
            out = activation(out)

            encoded_len = encoded_len // down_sample_rate
            # [B, 2 * C, T / down_sample_rate]
            out = down_sample_conv(inputs=out, input_len=encoded_len)

        out = self.rnn(inputs=out, input_len=encoded_len)
        out = self.post_activation(out)
        # [B, encoded_dim, T_encoded]
        encoded = self.post_conv(inputs=out, input_len=encoded_len)
        return encoded, encoded_len


class SEANetDecoder(NeuralModule):
    def __init__(
        self,
        up_sample_rates: Iterable[int] = (8, 5, 4, 2),
        base_channels: int = 512,
        in_kernel_size: int = 7,
        out_kernel_size: int = 3,
        encoded_dim: int = 128,
        activation: str = "elu",
        rnn_layers: int = 2,
        rnn_type: str = "lstm",
        rnn_skip: bool = True,
    ):
        assert in_kernel_size > 0
        assert out_kernel_size > 0

        super().__init__()

        self.up_sample_rates = up_sample_rates
        self.pre_conv = Conv1dNorm(in_channels=encoded_dim, out_channels=base_channels, kernel_size=in_kernel_size)
        self.rnn = SEANetRNN(dim=base_channels, num_layers=rnn_layers, rnn_type=rnn_type, use_skip=rnn_skip)

        in_channels = base_channels
        self.res_blocks = nn.ModuleList([])
        self.up_sample_conv_layers = nn.ModuleList([])
        self.activations = nn.ModuleList([])
        for i, up_sample_rate in enumerate(self.up_sample_rates):
            self.activations.append(CodecActivation(activation=activation, channels=in_channels))
            out_channels = in_channels // 2
            kernel_size = 2 * up_sample_rate
            up_sample_conv = ConvTranspose1dNorm(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=up_sample_rate,
            )
            in_channels = out_channels
            self.up_sample_conv_layers.append(up_sample_conv)

            res_block = SEANetResnetBlock(channels=in_channels)
            self.res_blocks.append(res_block)

        self.post_activation = CodecActivation(activation=activation, channels=in_channels)
        self.post_conv = Conv1dNorm(in_channels=in_channels, out_channels=1, kernel_size=out_kernel_size)
        self.out_activation = nn.Tanh()

    @property
    def input_types(self):
        return {
            "inputs": [NeuralType(('B', 'D', 'T_encoded'), EncodedRepresentation())],
            "input_len": [NeuralType(tuple('B'), LengthsType())],
        }

    @property
    def output_types(self):
        return {
            "audio": NeuralType(('B', 'T_audio'), AudioSignal()),
            "audio_len": NeuralType(tuple('B'), LengthsType()),
        }

    def remove_weight_norm(self):
        self.pre_conv.remove_weight_norm()
        for up_sample_conv in self.up_sample_conv_layers:
            up_sample_conv.remove_weight_norm()
        for res_block in self.res_blocks:
            res_block.remove_weight_norm()

    @typecheck()
    def forward(self, inputs, input_len):
        audio_len = input_len
        # [B, C, T_encoded]
        out = self.pre_conv(inputs=inputs, input_len=audio_len)
        out = self.rnn(inputs=out, input_len=audio_len)
        for res_block, up_sample_conv, up_sample_rate, activation in zip(
            self.res_blocks, self.up_sample_conv_layers, self.up_sample_rates, self.activations
        ):
            audio_len = audio_len * up_sample_rate
            out = activation(out)
            # [B, C / 2, T * up_sample_rate]
            out = up_sample_conv(inputs=out, input_len=audio_len)
            out = res_block(inputs=out, input_len=audio_len)

        out = self.post_activation(out)
        # [B, 1, T_audio]
        out = self.post_conv(inputs=out, input_len=audio_len)
        audio = self.out_activation(out)
        audio = rearrange(audio, "B 1 T -> B T")
        return audio, audio_len


class DiscriminatorSTFT(NeuralModule):
    def __init__(self, resolution, lrelu_slope=0.1):
        super().__init__()

        self.n_fft, self.hop_length, self.win_length = resolution
        self.register_buffer("window", torch.hann_window(self.win_length, periodic=False))
        self.activation = nn.LeakyReLU(lrelu_slope)

        self.conv_layers = nn.ModuleList(
            [
                Conv2dNorm(2, 32, kernel_size=(3, 9)),
                Conv2dNorm(32, 32, kernel_size=(3, 9), dilation=(1, 1), stride=(1, 2)),
                Conv2dNorm(32, 32, kernel_size=(3, 9), dilation=(2, 1), stride=(1, 2)),
                Conv2dNorm(32, 32, kernel_size=(3, 9), dilation=(4, 1), stride=(1, 2)),
                Conv2dNorm(32, 32, kernel_size=(3, 3)),
            ]
        )
        self.conv_post = Conv2dNorm(32, 1, kernel_size=(3, 3))

    def stft(self, audio):
        # [B, fft, T_spec]
        out = torch.stft(
            audio,
            n_fft=self.n_fft,
            hop_length=self.hop_length,
            win_length=self.win_length,
            window=self.window,
            normalized=True,
            center=True,
            return_complex=True,
        )
        out = rearrange(out, "B fft T -> B 1 T fft")
        # [batch, 2, T_spec, fft]
        out = torch.cat([out.real, out.imag], dim=1)
        return out

    @property
    def input_types(self):
        return {
            "audio": NeuralType(('B', 'T_audio'), AudioSignal()),
        }

    @property
    def output_types(self):
        return {
            "scores": NeuralType(('B', 'C', 'T_spec'), VoidType()),
            "fmap": [NeuralType(('B', 'D', 'T_spec', 'C'), VoidType())],
        }

    @typecheck()
    def forward(self, audio):
        fmap = []

        # [batch, 2, T_spec, fft]
        out = self.stft(audio)
        for conv in self.conv_layers:
            # [batch, filters, T_spec, fft // 2**i]
            out = conv(inputs=out)
            out = self.activation(out)
            fmap.append(out)
        # [batch, 1, T_spec, fft // 8]
        scores = self.conv_post(inputs=out)
        fmap.append(scores)
        scores = rearrange(scores, "B 1 T C -> B C T")

        return scores, fmap


class MultiResolutionDiscriminatorSTFT(NeuralModule):
    def __init__(self, resolutions):
        super().__init__()
        self.discriminators = nn.ModuleList([DiscriminatorSTFT(res) for res in resolutions])

    @property
    def input_types(self):
        return {
            "audio_real": NeuralType(('B', 'T_audio'), AudioSignal()),
            "audio_gen": NeuralType(('B', 'T_audio'), AudioSignal()),
        }

    @property
    def output_types(self):
        return {
            "scores_real": [NeuralType(('B', 'C', 'T_spec'), VoidType())],
            "scores_gen": [NeuralType(('B', 'C', 'T_spec'), VoidType())],
            "fmaps_real": [[NeuralType(('B', 'D', 'T_spec', 'C'), VoidType())]],
            "fmaps_gen": [[NeuralType(('B', 'D', 'T_spec', 'C'), VoidType())]],
        }

    @typecheck()
    def forward(self, audio_real, audio_gen):
        scores_real = []
        scores_gen = []
        fmaps_real = []
        fmaps_gen = []

        for disc in self.discriminators:
            score_real, fmap_real = disc(audio=audio_real)
            scores_real.append(score_real)
            fmaps_real.append(fmap_real)

            score_gen, fmap_gen = disc(audio=audio_gen)
            scores_gen.append(score_gen)
            fmaps_gen.append(fmap_gen)

        return scores_real, scores_gen, fmaps_real, fmaps_gen


def _ema_inplace(moving_avg: Tensor, new: Tensor, decay: float) -> None:
    moving_avg.data.mul_(decay).add_(new, alpha=(1 - decay))


def _laplace_smoothing(inputs: Tensor, n_categories: int, epsilon: float = 1e-5) -> Tensor:
    input_sum = inputs.sum()
    smoothed = (inputs + epsilon) / (input_sum + n_categories * epsilon)
    return input_sum * smoothed


def _compute_distances(input1: Tensor, input2: Tensor) -> Tensor:
    """
    Compute pairwise L2 distance between two input tensors

    Args:
        input1: [B, D] first tensor.
        input2: [N, D] second tensor.

    Returns:
        [(B, D)] tensor of distances.
    """
    input2 = rearrange(input2, "N D -> D N")
    distances = input1.pow(2).sum(1, keepdim=True) - (2 * input1 @ input2) + input2.pow(2).sum(0, keepdim=True)
    return distances


def _sample_vectors(samples: Tensor, num_sample: int) -> Tensor:
    """
    Randomly sample from the input batch.

    Args:
        samples: [B, D] tensor with features to sample.
        num_sample: Number of samples to draw.
            If the value is less than or equal to B, then the samples will be unique.
            If the value is greater than B, then samples will be drawn with replacement.

    Returns:
        Tensor with num_sample values randomly sampled from the input batch.
    """
    device = samples.device
    total_samples = samples.shape[0]

    if total_samples >= num_sample:
        indices = torch.randperm(total_samples, device=device)[:num_sample]
    else:
        indices = torch.randint(low=0, high=total_samples, size=(num_sample,), device=device)

    return samples[indices]


def _k_means(samples: Tensor, num_clusters: int, num_iters: int = 10) -> Tuple[Tensor, Tensor]:
    """
    K-means clustering algorithm.

    Args:
        samples: [B, D] tensor with features to cluster
        num_clusters: K, the number of clusters.
        num_iters: Number of iterations of K-means to run.

    Returns:
        [K, D] cluster means and [K] bins counting how many input samples belong to each cluster
    """
    assert num_iters > 0

    input_dim = samples.shape[1]
    # [K, D]
    means = _sample_vectors(samples=samples, num_sample=num_clusters)

    for _ in range(num_iters):
        # [B, K]
        dists = _compute_distances(samples, means)

        # [N]
        buckets = dists.min(dim=1).indices
        buckets_repeated = repeat(buckets, "B -> B D", D=input_dim)
        # [K]
        bin_counts = torch.bincount(buckets, minlength=num_clusters)
        bin_counts_expanded = rearrange(bin_counts, "K -> K ()")

        # [K, D]
        new_means = buckets.new_zeros(num_clusters, input_dim, dtype=samples.dtype)
        new_means.scatter_add_(dim=0, index=buckets_repeated, src=samples)
        new_means = new_means / torch.clamp(bin_counts_expanded, min=1)
        means = torch.where(bin_counts_expanded == 0, means, new_means)

    return means, bin_counts


def _mask_3d(tensor: Tensor, lengths: Tensor):
    """
    Mask 3d tensor with time on 1st axis.

    Args:
        tensor: tensor of shape (B, T, D)
        lengths: LongTensor of shape (B,)
    Returns:
        Masked Tensor (B, T, D)
    """
    batch_size, max_lengths, _ = tensor.shape
    mask = torch.ones(batch_size, max_lengths, 1).cumsum(dim=1).type_as(lengths)
    mask = mask <= rearrange(lengths, "b -> b 1 1")
    return tensor * mask


class EuclideanCodebook(NeuralModule):
    """
    Codebook with Euclidean distance.

    Args:
        codebook_size: Number of codes to use.
        codebook_dim: Dimension of each code.
        decay: Decay for exponential moving average over the codebooks.
        threshold_ema_dead_code: Threshold for dead code expiration.
            During every iteration, replace codes with exponential moving average cluster size less than threshold
            with randomly selected values from the current batch.
        kmeans_iters: Optional int, if provided codes will be initialized from the centroids learned from
            kmeans_iters iterations of k-means clustering on the first training batch.
    """

    def __init__(
        self,
        codebook_size: int,
        codebook_dim: int,
        decay: float = 0.99,
        threshold_ema_dead_code: Optional[float] = 2.0,
        kmeans_iters: Optional[int] = 50,
    ):
        super().__init__()
        self.decay = decay

        if kmeans_iters:
            codes = nn.init.kaiming_uniform_(torch.empty(codebook_size, codebook_dim))
        else:
            codes = torch.zeros(codebook_size, codebook_dim)

        self.codebook_size = codebook_size

        self.kmeans_iters = kmeans_iters
        self.threshold_ema_dead_code = threshold_ema_dead_code

        self.register_buffer("initialized", Tensor([not kmeans_iters]))
        self.register_buffer("cluster_size", torch.zeros(codebook_size))
        self.register_buffer("codes", codes)
        self.register_buffer("codes_avg", codes.clone())

    @torch.jit.ignore
    def _init_codes(self, data):
        if self.initialized:
            return

        codes, cluster_size = _k_means(samples=data, num_clusters=self.codebook_size, num_iters=self.kmeans_iters)
        self.codes.data.copy_(codes)
        self.codes_avg.data.copy_(codes.clone())
        self.cluster_size.data.copy_(cluster_size)
        self.initialized.data.copy_(Tensor([True]))
        broadcast_tensors(self.buffers())

    def _expire_codes(self, inputs: Tensor) -> None:
        if not self.threshold_ema_dead_code:
            return

        expired_codes = self.cluster_size < self.threshold_ema_dead_code
        if not torch.any(expired_codes):
            return

        samples = _sample_vectors(samples=inputs, num_sample=self.codebook_size)
        expired_codes = rearrange(expired_codes, "K -> K ()")
        modified_codes = torch.where(expired_codes, samples, self.codes)
        self.codes.data.copy_(modified_codes)

        broadcast_tensors(self.buffers())

    def _update_codes(self, inputs: Tensor, indices: Tensor) -> None:
        code_onehot = F.one_hot(indices, self.codebook_size).type(inputs.dtype)
        code_onehot = rearrange(code_onehot, "B N -> N B")
        # [N]
        code_counts = code_onehot.sum(1)
        _ema_inplace(moving_avg=self.cluster_size, new=code_counts, decay=self.decay)
        # [N, D]
        code_sum = code_onehot @ inputs
        _ema_inplace(moving_avg=self.codes_avg, new=code_sum, decay=self.decay)

        cluster_size_smoothed = _laplace_smoothing(self.cluster_size, n_categories=self.codebook_size)
        cluster_size_smoothed = rearrange(cluster_size_smoothed, "N -> N ()")
        codes_normalized = self.codes_avg / cluster_size_smoothed
        self.codes.data.copy_(codes_normalized)

    def _quantize(self, inputs: Tensor) -> Tensor:
        # [B, N]
        dist = _compute_distances(inputs, self.codes)
        # [B]
        indices = dist.min(dim=1).indices
        return indices

    def _dequantize(self, indices: Tensor) -> Tensor:
        # [B, D]
        dequantized = F.embedding(indices, self.codes)
        return dequantized

    @property
    def input_types(self):
        return {
            "inputs": NeuralType(('B', 'T', 'D'), EncodedRepresentation()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        return {
            "dequantized": NeuralType(('B', 'T', 'D'), EncodedRepresentation()),
            "indices": NeuralType(('B', 'T'), Index()),
        }

    @typecheck()
    def forward(self, inputs, input_len):
        input_flat = rearrange(inputs, "B T D -> (B T) D")
        self._init_codes(input_flat)
        # [(B T)]
        indices_flat = self._quantize(inputs=input_flat)
        # [B, T]
        indices = indices_flat.view(*inputs.shape[:-1])
        # [B, T, D]
        dequantized = self._dequantize(indices=indices)

        if self.training:
            # We do expiry of codes here because buffers are in sync and all the workers will make the same decision.
            self._expire_codes(inputs=input_flat)
            self._update_codes(inputs=input_flat, indices=indices_flat)

        dequantized = _mask_3d(dequantized, input_len)
        indices = mask_sequence_tensor(indices, input_len)
        return dequantized, indices

    @typecheck(
        input_types={
            "inputs": NeuralType(('B', 'T', 'D'), EncodedRepresentation()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        },
        output_types={"indices": NeuralType(('B', 'T'), Index())},
    )
    def encode(self, inputs, input_len):
        input_flat = rearrange(inputs, "B T D -> (B T) D")
        # [(B T)]
        indices_flat = self._quantize(inputs=input_flat)
        # [B, T]
        indices = indices_flat.view(*inputs.shape[:-1])
        indices = mask_sequence_tensor(indices, input_len)
        return indices

    @typecheck(
        input_types={
            "indices": NeuralType(('B', 'T'), Index()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        },
        output_types={"dequantized": NeuralType(('B', 'T', 'D'), EncodedRepresentation())},
    )
    def decode(self, indices, input_len):
        # [B, T, D]
        dequantized = self._dequantize(indices=indices)
        dequantized = _mask_3d(dequantized, input_len)
        return dequantized


class ResidualVectorQuantizer(VectorQuantizerBase):
    """
    Residual vector quantization (RVQ) algorithm as described in https://arxiv.org/pdf/2107.03312.pdf.

    Args:
        num_codebooks: Number of codebooks to use.
        codebook_size: Number of codes to use for each codebook.
        codebook_dim: Dimension of each code.
        decay: Decay for exponential moving average over the codebooks.
        threshold_ema_dead_code: Threshold for dead code expiration.
            During every iteration, replace codes with exponential moving average cluster size less than threshold
            with randomly selected values from the current batch.
        kmeans_iters: Optional int, if provided codes will be initialized from the centroids learned from
            kmeans_iters iterations of k-means clustering on the first training batch.
    """

    def __init__(
        self,
        num_codebooks: int,
        codebook_size: int = 1024,
        codebook_dim: int = 128,
        decay: float = 0.99,
        threshold_ema_dead_code: Optional[float] = 2.0,
        kmeans_iters: Optional[int] = 50,
    ):
        super().__init__()
        self.codebook_dim = codebook_dim
        self.commit_loss_fn = MaskedMSELoss()

        self.codebooks = nn.ModuleList(
            [
                EuclideanCodebook(
                    codebook_size=codebook_size,
                    codebook_dim=codebook_dim,
                    decay=decay,
                    threshold_ema_dead_code=threshold_ema_dead_code,
                    kmeans_iters=kmeans_iters,
                )
                for _ in range(num_codebooks)
            ]
        )

    @property
    def num_codebooks(self):
        """Returns the number of codebooks."""
        return len(self.codebooks)

    @property
    def codebook_size(self):
        """Returns the size of the codebook for each group."""
        return self.codebooks[0].codebook_size

    # Override output types, since this quantizer returns commit_loss
    @property
    def output_types(self):
        return {
            "dequantized": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
            "indices": NeuralType(('D', 'B', 'T'), Index()),
            "commit_loss": NeuralType((), LossType()),
        }

    @typecheck()
    def forward(self, inputs: Tensor, input_len: Tensor) -> Tuple[Tensor, Tensor, float]:
        commit_loss = 0.0
        residual = rearrange(inputs, "B D T -> B T D")

        index_list = []
        dequantized = torch.zeros_like(residual)
        for codebook in self.codebooks:
            dequantized_i, indices_i = codebook(inputs=residual, input_len=input_len)

            if self.training:
                dequantized_i_const = dequantized_i.detach()

                commit_loss_i = self.commit_loss_fn(
                    predicted=rearrange(residual, "B T D -> B D T"),
                    target=rearrange(dequantized_i_const, "B T D -> B D T"),
                    target_len=input_len,
                )
                commit_loss = commit_loss + commit_loss_i

                residual = residual - dequantized_i_const
                dequantized_i = residual + (dequantized_i - residual).detach()
            else:
                residual = residual - dequantized_i

            dequantized = dequantized + dequantized_i
            index_list.append(indices_i)

        # [N, B, T], first dimension is number of codebooks
        indices = torch.stack(index_list)
        dequantized = rearrange(dequantized, "B T D -> B D T")
        return dequantized, indices, commit_loss

    @typecheck(
        input_types={
            "inputs": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        },
        output_types={"indices": NeuralType(('D', 'B', 'T'), Index())},
    )
    def encode(self, inputs: Tensor, input_len: Tensor) -> Tensor:
        residual = rearrange(inputs, "B D T -> B T D")
        index_list = []
        for codebook in self.codebooks:
            # [B, T]
            indices_i = codebook.encode(inputs=residual, input_len=input_len)
            # [B, D, T]
            dequantized_i = codebook.decode(indices=indices_i, input_len=input_len)
            residual = residual - dequantized_i
            index_list.append(indices_i)
        # [N, B, T]
        indices = torch.stack(index_list)
        return indices

    @typecheck(
        input_types={
            "indices": NeuralType(('D', 'B', 'T'), Index()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        },
        output_types={
            "dequantized": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
        },
    )
    def decode(self, indices: Tensor, input_len: Tensor) -> Tensor:
        # [B, T, D]
        dequantized = torch.zeros([indices.shape[1], indices.shape[2], self.codebook_dim], device=indices.device)
        for codebook_indices, codebook in zip(indices, self.codebooks):
            dequantized_i = codebook.decode(indices=codebook_indices, input_len=input_len)
            dequantized = dequantized + dequantized_i
        dequantized = rearrange(dequantized, "B T D -> B D T")
        return dequantized


class GroupResidualVectorQuantizer(VectorQuantizerBase):
    """Split the input vector into groups and apply RVQ on each group separately.

    Args:
        num_codebooks: total number of codebooks
        num_groups: number of groups to split the input into, each group will be quantized separately using num_codebooks//num_groups codebooks
        codebook_dim: embedding dimension, will be split into num_groups
        **kwargs: parameters of ResidualVectorQuantizer

    References:
        Yang et al, HiFi-Codec: Group-residual Vector quantization for High Fidelity Audio Codec, 2023 (http://arxiv.org/abs/2305.02765).
    """

    def __init__(self, num_codebooks: int, num_groups: int, codebook_dim: int, **kwargs):
        super().__init__()

        self._num_codebooks = num_codebooks
        self.num_groups = num_groups
        self.codebook_dim = codebook_dim

        # Initialize RVQ for each group
        self.rvqs = torch.nn.ModuleList(
            [
                ResidualVectorQuantizer(
                    num_codebooks=self.num_codebooks_per_group, codebook_dim=self.codebook_dim_per_group, **kwargs
                )
                for _ in range(self.num_groups)
            ]
        )

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tnum_codebooks:           %d', self.num_codebooks)
        logging.debug('\tnum_groups:              %d', self.num_groups)
        logging.debug('\tcodebook_dim:            %d', self.codebook_dim)
        logging.debug('\tnum_codebooks_per_group: %d', self.num_codebooks_per_group)
        logging.debug('\tcodebook_dim_per_group:  %d', self.codebook_dim_per_group)

    @property
    def num_codebooks(self):
        """Returns the number of codebooks."""
        return self._num_codebooks

    @property
    def codebook_size(self):
        """Returns the size of the codebook for each group."""
        return self.rvqs[0].codebook_size

    @property
    def num_codebooks_per_group(self):
        """Number of codebooks for each group."""
        if self.num_codebooks % self.num_groups != 0:
            raise ValueError(
                f'num_codebooks ({self.num_codebooks}) must be divisible by num_groups ({self.num_groups})'
            )

        return self.num_codebooks // self.num_groups

    @property
    def codebook_dim_per_group(self):
        """Input vector dimension for each group."""
        if self.codebook_dim % self.num_groups != 0:
            raise ValueError(f'codebook_dim ({self.codebook_dim}) must be divisible by num_groups ({self.num_groups})')

        return self.codebook_dim // self.num_groups

    # Override output types, since this quantizer returns commit_loss
    @property
    def output_types(self):
        return {
            "dequantized": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
            "indices": NeuralType(('D', 'B', 'T'), Index()),
            "commit_loss": NeuralType((), LossType()),
        }

    @typecheck()
    def forward(self, inputs, input_len):
        """Quantize each group separately, then concatenate the results."""
        inputs_grouped = inputs.chunk(self.num_groups, dim=1)

        dequantized, indices = [], []
        commit_loss = 0

        for in_group, rvq_group in zip(inputs_grouped, self.rvqs):
            dequantized_group, indices_group, commit_loss_group = rvq_group(inputs=in_group, input_len=input_len)
            dequantized.append(dequantized_group)
            indices.append(indices_group)
            commit_loss += commit_loss_group

        # concatenate along the feature dimension
        dequantized = torch.cat(dequantized, dim=1)

        # concatente along the codebook dimension
        indices = torch.cat(indices, dim=0)

        return dequantized, indices, commit_loss

    @typecheck(
        input_types={
            "inputs": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        },
        output_types={"indices": NeuralType(('D', 'B', 'T'), Index())},
    )
    def encode(self, inputs: Tensor, input_len: Tensor) -> Tensor:
        """Input is split into groups, each group is encoded separately, then the results are concatenated."""
        inputs_grouped = inputs.chunk(self.num_groups, dim=1)
        indices = []

        for in_group, rvq_group in zip(inputs_grouped, self.rvqs):
            indices_group = rvq_group.encode(inputs=in_group, input_len=input_len)
            indices.append(indices_group)

        # concatenate along the codebook dimension
        indices = torch.cat(indices, dim=0)

        return indices

    @typecheck(
        input_types={
            "indices": NeuralType(('D', 'B', 'T'), Index()),
            "input_len": NeuralType(tuple('B'), LengthsType()),
        },
        output_types={
            "dequantized": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
        },
    )
    def decode(self, indices: Tensor, input_len: Tensor) -> Tensor:
        """Input indices are split into groups, each group is decoded separately, then the results are concatenated."""
        indices_grouped = indices.chunk(self.num_groups, dim=0)
        dequantized = []

        for indices_group, rvq_group in zip(indices_grouped, self.rvqs):
            dequantized_group = rvq_group.decode(indices=indices_group, input_len=input_len)
            dequantized.append(dequantized_group)

        # concatenate along the feature dimension
        dequantized = torch.cat(dequantized, dim=1)

        return dequantized



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (c) 2018 Ryan Leary
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# This file contains code artifacts adapted from https://github.com/ryanleary/patter
import copy
import inspect
import io
import os
import random
import subprocess
from tempfile import NamedTemporaryFile
from typing import Any, List, Optional, Union

import librosa
import numpy as np
import soundfile as sf
from scipy import signal

from nemo.collections.asr.parts.preprocessing.segment import AudioSegment
from nemo.collections.common.parts.preprocessing import collections, parsers
from nemo.core.classes import IterableDataset
from nemo.utils import logging

# TODO @blisc: Perhaps refactor instead of import guarding
HAVE_OMEGACONG_WEBDATASET = True
try:
    from omegaconf import DictConfig, OmegaConf

    from nemo.utils import webdataset as wds

except ModuleNotFoundError:
    from nemo.utils.exceptions import LightningNotInstalledException

    HAVE_OMEGACONG_WEBDATASET = False


try:
    from nemo.collections.asr.parts.utils import numba_utils

    HAVE_NUMBA = True
except (ImportError, ModuleNotFoundError):
    HAVE_NUMBA = False


def read_one_audiosegment(manifest, target_sr, tarred_audio=False, audio_dataset=None):
    if tarred_audio:
        if audio_dataset is None:
            raise TypeError("Expected augmentation dataset but got None")
        audio_file, file_id, manifest_entry = next(audio_dataset)

        offset = 0 if manifest_entry.offset is None else manifest_entry.offset
        duration = 0 if manifest_entry.duration is None else manifest_entry.duration

    else:
        audio_record = random.sample(manifest.data, 1)[0]
        audio_file = audio_record.audio_file
        offset = 0 if audio_record.offset is None else audio_record.offset
        duration = 0 if audio_record.duration is None else audio_record.duration

    return AudioSegment.from_file(audio_file, target_sr=target_sr, offset=offset, duration=duration)


class Perturbation(object):
    def max_augmentation_length(self, length):
        return length

    def perturb(self, data):
        raise NotImplementedError


class SpeedPerturbation(Perturbation):
    """
    Performs Speed Augmentation by re-sampling the data to a different sampling rate,
    which does not preserve pitch.

    Note: This is a very slow operation for online augmentation. If space allows,
    it is preferable to pre-compute and save the files to augment the dataset.

    Args:
        sr: Original sampling rate.
        resample_type: Type of resampling operation that will be performed.
            For better speed using `resampy`'s fast resampling method, use `resample_type='kaiser_fast'`.
            For high-quality resampling, set `resample_type='kaiser_best'`.
            To use `scipy.signal.resample`, set `resample_type='fft'` or `resample_type='scipy'`
        min_speed_rate: Minimum sampling rate modifier.
        max_speed_rate: Maximum sampling rate modifier.
        num_rates: Number of discrete rates to allow. Can be a positive or negative
            integer.
            If a positive integer greater than 0 is provided, the range of
            speed rates will be discretized into `num_rates` values.
            If a negative integer or 0 is provided, the full range of speed rates
            will be sampled uniformly.
            Note: If a positive integer is provided and the resultant discretized
            range of rates contains the value '1.0', then those samples with rate=1.0,
            will not be augmented at all and simply skipped. This is to unnecessary
            augmentation and increase computation time. Effective augmentation chance
            in such a case is = `prob * (num_rates - 1 / num_rates) * 100`% chance
            where `prob` is the global probability of a sample being augmented.
        rng: Random seed. Default is None
    """

    def __init__(self, sr, resample_type, min_speed_rate=0.9, max_speed_rate=1.1, num_rates=5, rng=None):

        min_rate = min(min_speed_rate, max_speed_rate)
        if min_rate < 0.0:
            raise ValueError("Minimum sampling rate modifier must be > 0.")

        if resample_type not in ('kaiser_best', 'kaiser_fast', 'fft', 'scipy'):
            raise ValueError("Supported `resample_type` values are ('kaiser_best', 'kaiser_fast', 'fft', 'scipy')")

        self._sr = sr
        self._min_rate = min_speed_rate
        self._max_rate = max_speed_rate
        self._num_rates = num_rates
        if num_rates > 0:
            self._rates = np.linspace(self._min_rate, self._max_rate, self._num_rates, endpoint=True)
        self._res_type = resample_type
        random.seed(rng) if rng else None

    def max_augmentation_length(self, length):
        return length * self._max_rate

    def perturb(self, data):
        # Select speed rate either from choice or random sample
        if self._num_rates < 0:
            speed_rate = random.uniform(self._min_rate, self._max_rate)
        else:
            speed_rate = random.choice(self._rates)

        # Skip perturbation in case of identity speed rate
        if speed_rate == 1.0:
            return

        new_sr = int(self._sr * speed_rate)
        try:
            data._samples = librosa.core.resample(
                data._samples, orig_sr=self._sr, target_sr=new_sr, res_type=self._res_type
            )
        except Exception as e:
            logging.warning(f"Failed to resample audio from {self._sr} to {new_sr}. Skipping augmentation. Error: {e}")
            return


class TimeStretchPerturbation(Perturbation):
    """
    Time-stretch an audio series by a fixed rate while preserving pitch, based on [1]_, [2]_.

    Note:
    This is a simplified implementation, intended primarily for reference and pedagogical purposes.
    It makes no attempt to handle transients, and is likely to produce audible artifacts.

    References
    ----------
    .. [1] Ellis, D. P. W. "A phase vocoder in Matlab." Columbia University, 2002.
       `<http://www.ee.columbia.edu/~dpwe/resources/matlab/pvoc/>`_
    .. [2] librosa.effects.time_stretch
       `<https://librosa.org/doc/main/generated/librosa.effects.time_stretch.html>`_

    Args:
        min_speed_rate: Minimum sampling rate modifier.
        max_speed_rate: Maximum sampling rate modifier.
        num_rates: Number of discrete rates to allow. Can be a positive or negative
            integer.
            If a positive integer greater than 0 is provided, the range of
            speed rates will be discretized into `num_rates` values.
            If a negative integer or 0 is provided, the full range of speed rates
            will be sampled uniformly.
            Note: If a positive integer is provided and the resultant discretized
            range of rates contains the value '1.0', then those samples with rate=1.0,
            will not be augmented at all and simply skipped. This is to avoid unnecessary
            augmentation and increase computation time. Effective augmentation chance
            in such a case is = `prob * (num_rates - 1 / num_rates) * 100`% chance
            where `prob` is the global probability of a sample being augmented.
        n_fft: Number of fft filters to be computed.
        rng: Random seed. Default is None
    """

    def __init__(self, min_speed_rate=0.9, max_speed_rate=1.1, num_rates=5, n_fft=512, rng=None):

        min_rate = min(min_speed_rate, max_speed_rate)
        if min_rate < 0.0:
            raise ValueError("Minimum sampling rate modifier must be > 0.")

        self._min_rate = min_speed_rate
        self._max_rate = max_speed_rate
        self._num_rates = num_rates
        if num_rates > 0:
            self._rates = np.linspace(self._min_rate, self._max_rate, self._num_rates, endpoint=True)
        random.seed(rng) if rng else None

        # Pre-compute constants
        self._n_fft = int(n_fft)
        self._hop_length = int(n_fft // 2)

        # Pre-allocate buffers
        self._phi_advance_fast = np.linspace(0, np.pi * self._hop_length, self._hop_length + 1)
        self._scale_buffer_fast = np.empty(self._hop_length + 1, dtype=np.float32)

        self._phi_advance_slow = np.linspace(0, np.pi * self._n_fft, self._n_fft + 1)
        self._scale_buffer_slow = np.empty(self._n_fft + 1, dtype=np.float32)

    def max_augmentation_length(self, length):
        return length * self._max_rate

    def perturb(self, data):
        # Select speed rate either from choice or random sample
        if self._num_rates < 0:
            speed_rate = random.uniform(self._min_rate, self._max_rate)
        else:
            speed_rate = random.choice(self._rates)

        # Skip perturbation in case of identity speed rate
        if speed_rate == 1.0:
            return

        # Increase `n_fft` based on task (speed up or slow down audio)
        # This greatly reduces upper bound of maximum time taken
        # to compute slowed down audio segments.
        if speed_rate >= 1.0:  # Speed up audio
            fft_multiplier = 1
            phi_advance = self._phi_advance_fast
            scale_buffer = self._scale_buffer_fast

        else:  # Slow down audio
            fft_multiplier = 2
            phi_advance = self._phi_advance_slow
            scale_buffer = self._scale_buffer_slow

        n_fft = int(self._n_fft * fft_multiplier)
        hop_length = int(self._hop_length * fft_multiplier)

        # Perform short-term Fourier transform (STFT)
        stft = librosa.core.stft(data._samples, n_fft=n_fft, hop_length=hop_length)

        # Stretch by phase vocoding
        if HAVE_NUMBA:
            stft_stretch = numba_utils.phase_vocoder(stft, speed_rate, phi_advance, scale_buffer)

        else:
            stft_stretch = librosa.core.phase_vocoder(stft, speed_rate, hop_length)

        # Predict the length of y_stretch
        len_stretch = int(round(len(data._samples) / speed_rate))

        # Invert the STFT
        y_stretch = librosa.core.istft(
            stft_stretch, dtype=data._samples.dtype, hop_length=hop_length, length=len_stretch
        )

        data._samples = y_stretch


class SilencePerturbation(Perturbation):
    """
    Applies random silence at the start and/or end of the audio.

    Args:
        min_start_silence_secs (float): Min start silence level in secs
        max_start_silence_secs (float): Max start silence level in secs
        min_end_silence_secs (float): Min end silence level in secs
        max_end_silence_secs (float): Max end silence level in secs
        rng (int): Random seed. Default is None
        value: (float): value representing silence to be added to audio array.
    """

    def __init__(
        self,
        min_start_silence_secs: float = 0,
        max_start_silence_secs: float = 0,
        min_end_silence_secs: float = 0,
        max_end_silence_secs: float = 0,
        rng: int = None,
        value: float = 0,
    ):
        self._min_start_silence_secs = min_start_silence_secs
        self._max_start_silence_secs = max_start_silence_secs
        self._min_end_silence_secs = min_end_silence_secs
        self._max_end_silence_secs = max_end_silence_secs

        random.seed(rng) if rng else None
        self._value = value

    def perturb(self, data):
        start_silence_len = random.uniform(self._min_start_silence_secs, self._max_start_silence_secs)
        end_silence_len = random.uniform(self._min_end_silence_secs, self._max_end_silence_secs)
        start = np.full((int(start_silence_len * data.sample_rate),), self._value)
        end = np.full((int(end_silence_len * data.sample_rate),), self._value)

        data._samples = np.concatenate([start, data._samples, end])


class GainPerturbation(Perturbation):
    """
    Applies random gain to the audio.

    Args:
        min_gain_dbfs (float): Min gain level in dB
        max_gain_dbfs (float): Max gain level in dB
        rng (int): Random seed. Default is None
    """

    def __init__(self, min_gain_dbfs=-10, max_gain_dbfs=10, rng=None):
        self._min_gain_dbfs = min_gain_dbfs
        self._max_gain_dbfs = max_gain_dbfs
        random.seed(rng) if rng else None

    def perturb(self, data):
        gain = random.uniform(self._min_gain_dbfs, self._max_gain_dbfs)
        data._samples = data._samples * (10.0 ** (gain / 20.0))


class ImpulsePerturbation(Perturbation):
    """
    Convolves audio with a Room Impulse Response.

    Args:
        manifest_path (list): Manifest file for RIRs
        audio_tar_filepaths (list): Tar files, if RIR audio files are tarred
        shuffle_n (int): Shuffle parameter for shuffling buffered files from the tar files
        normalize_impulse (bool): Normalize impulse response to zero mean and amplitude 1
        shift_impulse (bool): Shift impulse response to adjust for delay at the beginning
        rng (int): Random seed. Default is None
    """

    def __init__(
        self,
        manifest_path=None,
        audio_tar_filepaths=None,
        shuffle_n=128,
        normalize_impulse=False,
        shift_impulse=False,
        rng=None,
    ):
        self._manifest = collections.ASRAudioText(manifest_path, parser=parsers.make_parser([]), index_by_file_id=True)
        self._audiodataset = None
        self._tarred_audio = False
        self._normalize_impulse = normalize_impulse
        self._shift_impulse = shift_impulse
        self._data_iterator = None

        if audio_tar_filepaths:
            self._tarred_audio = True
            self._audiodataset = AugmentationDataset(manifest_path, audio_tar_filepaths, shuffle_n)
            self._data_iterator = iter(self._audiodataset)

        self._rng = rng
        random.seed(self._rng) if rng else None

    def perturb(self, data):
        impulse = read_one_audiosegment(
            self._manifest,
            data.sample_rate,
            tarred_audio=self._tarred_audio,
            audio_dataset=self._data_iterator,
        )

        # normalize if necessary
        if self._normalize_impulse:
            # normalize the impulse response to zero mean and amplitude 1
            impulse_norm = impulse.samples - np.mean(impulse.samples)
            impulse_norm /= max(abs(impulse_norm))
        else:
            impulse_norm = impulse.samples

        # len of input data samples
        len_data = len(data._samples)

        if max(abs(data._samples)) == 0:
            logging.warning("Zero audio input found, skipping impulse perturbation.")
            return

        # convolve with the full impulse response
        data._samples = signal.fftconvolve(data._samples, impulse_norm, "full")

        # compensate the dominant path propagation delay
        if self._shift_impulse:
            # Find the peak of the IR and shift the output to the left
            max_ind = np.argmax(np.abs(impulse_norm))
            data._samples = data._samples[max_ind:]

        # trim to match the input data length
        data._samples = data._samples[:len_data]

        if max(abs(data._samples)) == 0:
            logging.warning("Zero audio input found after impulse perturbation.")
            return

        # normalize data samples to [-1,1] after rir convolution to avoid nans with fp16 training
        data._samples = data._samples / max(abs(data._samples))


class ShiftPerturbation(Perturbation):
    """
    Perturbs audio by shifting the audio in time by a random amount between min_shift_ms and max_shift_ms.
    The final length of the audio is kept unaltered by padding the audio with zeros.


    Args:
        min_shift_ms (float): Minimum time in milliseconds by which audio will be shifted
        max_shift_ms (float): Maximum time in milliseconds by which audio will be shifted
        rng (int): Random seed. Default is None
    """

    def __init__(self, min_shift_ms=-5.0, max_shift_ms=5.0, rng=None):
        self._min_shift_ms = min_shift_ms
        self._max_shift_ms = max_shift_ms
        random.seed(rng) if rng else None

    def perturb(self, data):
        shift_ms = random.uniform(self._min_shift_ms, self._max_shift_ms)
        if abs(shift_ms) / 1000 > data.duration:
            # TODO: do something smarter than just ignore this condition
            return
        shift_samples = int(shift_ms * data.sample_rate // 1000)
        # logging.debug("shift: %s", shift_samples)
        if shift_samples < 0:
            data._samples[-shift_samples:] = data._samples[:shift_samples]
            data._samples[:-shift_samples] = 0
        elif shift_samples > 0:
            data._samples[:-shift_samples] = data._samples[shift_samples:]
            data._samples[-shift_samples:] = 0


class NoisePerturbation(Perturbation):
    """
    Perturbation that adds noise to input audio.

    Args:
        manifest_path (str): Manifest file with paths to noise files
        min_snr_db (float): Minimum SNR of audio after noise is added
        max_snr_db (float): Maximum SNR of audio after noise is added
        max_gain_db (float): Maximum gain that can be applied on the noise sample
        audio_tar_filepaths (list) : Tar files, if noise audio files are tarred
        shuffle_n (int): Shuffle parameter for shuffling buffered files from the tar files
        orig_sr (int): Original sampling rate of the noise files
        rng (int): Random seed. Default is None
    """

    def __init__(
        self,
        manifest_path=None,
        min_snr_db=10,
        max_snr_db=50,
        max_gain_db=300.0,
        rng=None,
        audio_tar_filepaths=None,
        shuffle_n=100,
        orig_sr=16000,
    ):
        self._manifest = collections.ASRAudioText(manifest_path, parser=parsers.make_parser([]), index_by_file_id=True)
        self._audiodataset = None
        self._tarred_audio = False
        self._orig_sr = orig_sr
        self._data_iterator = None

        if audio_tar_filepaths:
            self._tarred_audio = True
            self._audiodataset = AugmentationDataset(manifest_path, audio_tar_filepaths, shuffle_n)
            self._data_iterator = iter(self._audiodataset)

        random.seed(rng) if rng else None
        self._rng = rng

        self._min_snr_db = min_snr_db
        self._max_snr_db = max_snr_db
        self._max_gain_db = max_gain_db

    @property
    def orig_sr(self):
        return self._orig_sr

    def get_one_noise_sample(self, target_sr):
        return read_one_audiosegment(
            self._manifest, target_sr, tarred_audio=self._tarred_audio, audio_dataset=self._data_iterator
        )

    def perturb(self, data, ref_mic=0):
        """
        Args:
            data (AudioSegment): audio data
            ref_mic (int): reference mic index for scaling multi-channel audios
        """
        noise = read_one_audiosegment(
            self._manifest,
            data.sample_rate,
            tarred_audio=self._tarred_audio,
            audio_dataset=self._data_iterator,
        )
        self.perturb_with_input_noise(data, noise, ref_mic=ref_mic)

    def perturb_with_input_noise(self, data, noise, data_rms=None, ref_mic=0):
        """
        Args:
            data (AudioSegment): audio data
            noise (AudioSegment): noise data
            data_rms (Union[float, List[float]): rms_db for data input
            ref_mic (int): reference mic index for scaling multi-channel audios
        """
        if data.num_channels != noise.num_channels:
            raise ValueError(
                f"Found mismatched channels for data ({data.num_channels}) and noise ({noise.num_channels})."
            )

        if not (0 <= ref_mic < data.num_channels):
            raise ValueError(
                f" reference mic ID must be an integer in [0, {data.num_channels}), got {ref_mic} instead."
            )

        snr_db = random.uniform(self._min_snr_db, self._max_snr_db)

        if data.is_empty():
            logging.warning(
                f"Empty audio segment found for {data.audio_file} with offset {data.offset} and duration {data.duration}."
            )

        if data_rms is None:
            data_rms = data.rms_db

        if noise.is_empty():
            logging.warning(
                f"Empty noise segment found for {noise.audio_file} with offset {noise.offset} and duration {noise.duration}."
            )
            noise_rms = -float("inf")
        else:
            noise_rms = noise.rms_db

        if data.is_empty() and noise.is_empty():
            logging.warning("Both data and noise segments are empty. Skipping perturbation.")
            return

        if data.num_channels > 1:
            noise_gain_db = data_rms[ref_mic] - noise_rms[ref_mic] - snr_db
        else:
            noise_gain_db = data_rms - noise_rms - snr_db
        noise_gain_db = min(noise_gain_db, self._max_gain_db)

        # calculate noise segment to use
        start_time = random.uniform(0.0, noise.duration - data.duration)
        if noise.duration > (start_time + data.duration):
            noise.subsegment(start_time=start_time, end_time=start_time + data.duration)

        # adjust gain for snr purposes and superimpose
        noise.gain_db(noise_gain_db)

        if noise._samples.shape[0] < data._samples.shape[0]:
            noise_idx = random.randint(0, data._samples.shape[0] - noise._samples.shape[0])
            data._samples[noise_idx : noise_idx + noise._samples.shape[0]] += noise._samples

        else:
            data._samples += noise._samples

    def perturb_with_foreground_noise(self, data, noise, data_rms=None, max_noise_dur=2, max_additions=1, ref_mic=0):
        """
        Args:
            data (AudioSegment): audio data
            noise (AudioSegment): noise data
            data_rms (Union[float, List[float]): rms_db for data input
            max_noise_dur: (float): max noise duration
            max_additions (int): number of times for adding noise
            ref_mic (int): reference mic index for scaling multi-channel audios
        """
        if data.num_channels != noise.num_channels:
            raise ValueError(
                f"Found mismatched channels for data ({data.num_channels}) and noise ({noise.num_channels})."
            )

        if not (0 <= ref_mic < data.num_channels):
            raise ValueError(
                f" reference mic ID must be an integer in [0, {data.num_channels}), got {ref_mic} instead."
            )

        snr_db = random.uniform(self._min_snr_db, self._max_snr_db)
        if not data_rms:
            data_rms = data.rms_db

        if data.num_channels > 1:
            noise_gain_db = data_rms[ref_mic] - noise.rms_db[ref_mic] - snr_db
        else:
            noise_gain_db = data_rms - noise.rms_db - snr_db
        noise_gain_db = min(noise_gain_db, self._max_gain_db)

        n_additions = random.randint(1, max_additions)

        for i in range(n_additions):
            noise_dur = random.uniform(0.0, max_noise_dur)
            start_time = random.uniform(0.0, noise.duration)
            start_sample = int(round(start_time * noise.sample_rate))
            end_sample = int(round(min(noise.duration, (start_time + noise_dur)) * noise.sample_rate))
            noise_samples = np.copy(noise._samples[start_sample:end_sample])
            # adjust gain for snr purposes and superimpose
            noise_samples *= 10.0 ** (noise_gain_db / 20.0)

            if noise_samples.shape[0] > data._samples.shape[0]:
                noise_samples = noise_samples[0 : data._samples.shape[0]]

            noise_idx = random.randint(0, data._samples.shape[0] - noise_samples.shape[0])
            data._samples[noise_idx : noise_idx + noise_samples.shape[0]] += noise_samples


class NoisePerturbationWithNormalization(Perturbation):
    """
    Perturbation that adds noise to input audio, with normalisation to specific decibel level.
    Also tiles shorter noise samples up to their corresponding clean audio length.

    Args:
        manifest_path (str or list): Manifest file with paths to noise files, can be list if using multiple noise sources
        min_snr_db (float): Minimum SNR of audio after noise is added
        max_snr_db (float): Maximum SNR of audio after noise is added
        snr_samples (list): A discrete list of SNRs DBs to sample from when mixing, will be used instead of [min_snr_db,max_snr_db]
        norm_to_db (float): Will normalise clean, noise, and mixed samples to this DB
        audio_tar_filepaths (str or list) : Tar files, if noise audio files are tarred, can be list for multiple sources
        shuffle_n (int): Shuffle parameter for shuffling buffered files from the tar files
        orig_sr (int): Original sampling rate of the noise files
        rng (int): Random seed. Default is None
        shard_strategy (str): if you're using tarred audio and wish to scatter instead of replicate, set this to 'scatter'
        epsilon (float): minimum value for RMS DB normalisation to avoid divide by zero
    """

    def __init__(
        self,
        manifest_path=None,
        min_snr_db=10,
        max_snr_db=50,
        snr_samples=None,
        norm_to_db=None,
        rng=None,
        audio_tar_filepaths=None,
        shuffle_n=128,
        orig_sr=16000,
        global_rank=0,
        world_size=1,
        shard_strategy='replicate',
        epsilon=0.01,
    ):
        # import here to avoid circular import error
        from nemo.collections.asr.data.audio_to_text import RandomizedChainDataset

        self._manifest = collections.ASRAudioText(manifest_path, parser=parsers.make_parser([]), index_by_file_id=True)
        self._audiodataset = None
        self._tarred_audio = False
        self._orig_sr = orig_sr
        self._data_iterator = None

        random.seed(rng) if rng else None
        self._rng = rng

        if audio_tar_filepaths:
            self._tarred_audio = True
            if isinstance(manifest_path, str):
                manifest_path = [manifest_path]
            if isinstance(audio_tar_filepaths, str):
                audio_tar_filepaths = [audio_tar_filepaths]
            datasets = []
            for tarred_audio_filepath, manifest_filepath in zip(audio_tar_filepaths, manifest_path):
                dataset = AugmentationDataset(
                    manifest_filepath,
                    tarred_audio_filepath,
                    shuffle_n,
                    rank=global_rank,
                    world_size=world_size,
                    shard_strategy=shard_strategy,
                )
                datasets.append(dataset)
            self._audiodataset = RandomizedChainDataset(
                datasets, rnd_seed=(rng if rng else random.randint(0, 30000)) + global_rank
            )
            if len(self._audiodataset) == 0:
                raise RuntimeError(
                    "NoisePerturbationWithNormalization detected a zero length RandomizedChainDataset, should never happen"
                )
            self._data_iterator = iter(self._audiodataset)

        self._min_snr_db = min_snr_db
        self._max_snr_db = max_snr_db
        self._norm_to_db = norm_to_db
        self._snr_samples = snr_samples if isinstance(snr_samples, list) and len(snr_samples) > 0 else None
        self._epsilon = epsilon

    @property
    def orig_sr(self):
        return self._orig_sr

    def read_one_audiosegment(self, target_sr):
        if self._tarred_audio:
            if self._data_iterator is None:
                raise TypeError("Expected valid iterator but got None")
            try:
                audio_file, file_id, manifest_entry = next(self._data_iterator)
            except StopIteration:
                self._data_iterator = iter(self._audiodataset)
                audio_file, file_id, manifest_entry = next(self._data_iterator)

            offset = 0 if manifest_entry.offset is None else manifest_entry.offset
            duration = 0 if manifest_entry.duration is None else manifest_entry.duration

        else:
            audio_record = random.sample(self._manifest.data, 1)[0]
            audio_file = audio_record.audio_file
            offset = 0 if audio_record.offset is None else audio_record.offset
            duration = 0 if audio_record.duration is None else audio_record.duration

        return AudioSegment.from_file(audio_file, target_sr=target_sr, offset=offset, duration=duration)

    def perturb(self, data, ref_mic=0):
        """
        Args:
            data (AudioSegment): audio data
            ref_mic (int): reference mic index for scaling multi-channel audios
        """

        noise = self.read_one_audiosegment(data.sample_rate)

        # noise samples need to be at least 1 second long to avoid strange oddities
        # in the RMS SNR mixing, so we have a fail-safe here to ensure at least 1 sec duration
        while noise.duration < 1:
            noise = self.read_one_audiosegment(data.sample_rate)

        self.perturb_with_input_noise(data, noise, ref_mic=ref_mic, norm_to_db=self._norm_to_db)

    def snr_mixer(self, clean, noise, snr, norm_to_db=-25.0):
        """
        Mixes the clean audio with the noise
        Args:
            clean (numpy array): the clean audio data
            noise (numpy array): the noise audio data
            snr (float): the SNR value for the mixing
            norm_to_db (float): the DB value to normalise to before mixing
        """
        clean = self.norm_audio_to_db(clean, norm_to_db)
        noise = self.norm_audio_to_db(noise, norm_to_db)

        # Set the noise level for a given SNR
        # note that if your noise doesn't overlap with your audio then your target SNR
        # may not be achievable. Consider using an rms-threshold in the future
        noisescalar = 10 ** (-snr / 20.0)
        noisenewlevel = noise * noisescalar
        noisyspeech = clean + noisenewlevel

        return clean, noisenewlevel, noisyspeech

    def norm_audio_to_db(self, x, norm_to_db):
        """
        Normalises audio signal to particular db, with some epsilon in-case of divide by zero
        Args:
            x (numpy array): input audio signal
            norm_to_db (float): the db to normalise to
        """
        rms = (x**2).mean(axis=0) ** 0.5
        rms = np.where(np.isclose(rms, 0), self._epsilon, rms)
        scalar = 10 ** (norm_to_db / 20.0) / rms
        return x * scalar

    def concatenate_noise_sample(self, clean, noise, fs, silence_length=0.25):
        """
        Tiles the noise array to match the clean audio array, with small silence between the joins
        Args:
            clean (numpy array): clean audio data
            noise (numpy array): noise audio data
            fs (int): sample rate used by both clean and noise audio data
            silence_length (float): the amount of silence (in secs) to insert before tiling
        """
        while len(noise) < len(clean):
            if noise.ndim > 1:
                zeros = np.zeros((int(fs * silence_length), noise.shape[-1]))
            else:
                zeros = np.zeros((int(fs * silence_length),))
            noiseconcat = np.append(noise, zeros, axis=0)
            noise = np.append(noiseconcat, noise, axis=0)

        return noise

    def perturb_with_input_noise(self, data, noise, data_rms=None, ref_mic=0, norm_to_db=-25.0):
        """
        Args:
            data (AudioSegment): audio data
            noise (AudioSegment): noise data
            data_rms (Union[float, List[float]): rms_db for data input
            ref_mic (int): reference mic index for scaling multi-channel audio, if set to None then
                           each channel will be scaled independently
            norm_to_db (float): will normalise all audio to this DB
        """
        if data.num_channels != noise.num_channels:
            raise ValueError(
                f"Found mismatched channels for data ({data.num_channels}) and noise ({noise.num_channels})."
            )

        if not (0 <= ref_mic < data.num_channels):
            raise ValueError(
                f" reference mic ID must be an integer in [0, {data.num_channels}), got {ref_mic} instead."
            )

        if self._snr_samples:
            snr_db = random.sample(self._snr_samples, 1)[0]
        else:
            snr_db = random.uniform(self._min_snr_db, self._max_snr_db)
        if data_rms is None:
            data_rms = data.rms_db[ref_mic] if isinstance(data.rms_db, (list, np.ndarray)) else data.rms_db

        if norm_to_db is None:
            norm_to_db = data_rms

        data_norm = data._samples
        noise_norm = noise._samples

        if len(data_norm) == 0:
            return

        if len(noise_norm) < len(data_norm):
            noise_norm = self.concatenate_noise_sample(data_norm, noise_norm, data.sample_rate)
        noise_norm = noise_norm[0 : len(data_norm)]

        _, _, noisy_snr = self.snr_mixer(clean=data_norm, noise=noise_norm, snr=snr_db, norm_to_db=norm_to_db)

        data._samples = noisy_snr


class WhiteNoisePerturbation(Perturbation):
    """
    Perturbation that adds white noise to an audio file in the training dataset.

    Args:
        min_level (int): Minimum level in dB at which white noise should be added
        max_level (int): Maximum level in dB at which white noise should be added
        rng (int): Random seed. Default is None
    """

    def __init__(self, min_level=-90, max_level=-46, rng=None):
        self.min_level = int(min_level)
        self.max_level = int(max_level)
        np.random.seed(rng) if rng else None

    def perturb(self, data):
        noise_level_db = np.random.randint(self.min_level, self.max_level, dtype='int32')
        noise_signal = np.random.randn(data._samples.shape[0]) * (10.0 ** (noise_level_db / 20.0))
        data._samples += noise_signal


class RirAndNoisePerturbation(Perturbation):
    """
    RIR augmentation with additive foreground and background noise.
    In this implementation audio data is augmented by first convolving the audio with a Room Impulse Response
    and then adding foreground noise and background noise at various SNRs. RIR, foreground and background noises
    should either be supplied with a manifest file or as tarred audio files (faster).

    Different sets of noise audio files based on the original sampling rate of the noise. This is useful while
    training a mixed sample rate model. For example, when training a mixed model with 8 kHz and 16 kHz audio with a
    target sampling rate of 16 kHz, one would want to augment 8 kHz data with 8 kHz noise rather than 16 kHz noise.

    Args:
        rir_manifest_path: Manifest file for RIRs
        rir_tar_filepaths: Tar files, if RIR audio files are tarred
        rir_prob: Probability of applying a RIR
        noise_manifest_paths: Foreground noise manifest path
        min_snr_db: Min SNR for foreground noise
        max_snr_db: Max SNR for background noise,
        noise_tar_filepaths: Tar files, if noise files are tarred
        apply_noise_rir: Whether to convolve foreground noise with a a random RIR
        orig_sample_rate: Original sampling rate of foreground noise audio
        max_additions: Max number of times foreground noise is added to an utterance,
        max_duration: Max duration of foreground noise
        bg_noise_manifest_paths: Background noise manifest path
        bg_min_snr_db: Min SNR for background noise
        bg_max_snr_db: Max SNR for background noise
        bg_noise_tar_filepaths: Tar files, if noise files are tarred
        bg_orig_sample_rate: Original sampling rate of background noise audio
        rng: Random seed. Default is None

    """

    def __init__(
        self,
        rir_manifest_path=None,
        rir_prob=0.5,
        noise_manifest_paths=None,
        noise_prob=1.0,
        min_snr_db=0,
        max_snr_db=50,
        rir_tar_filepaths=None,
        rir_shuffle_n=100,
        noise_tar_filepaths=None,
        apply_noise_rir=False,
        orig_sample_rate=None,
        max_additions=5,
        max_duration=2.0,
        bg_noise_manifest_paths=None,
        bg_noise_prob=1.0,
        bg_min_snr_db=10,
        bg_max_snr_db=50,
        bg_noise_tar_filepaths=None,
        bg_orig_sample_rate=None,
        rng=None,
    ):

        self._rir_prob = rir_prob
        self._noise_prob = noise_prob
        self._bg_noise_prob = bg_noise_prob
        random.seed(rng) if rng else None
        self._rir_perturber = ImpulsePerturbation(
            manifest_path=rir_manifest_path,
            audio_tar_filepaths=rir_tar_filepaths,
            shuffle_n=rir_shuffle_n,
            shift_impulse=True,
        )
        self._fg_noise_perturbers = None
        self._bg_noise_perturbers = None
        if noise_manifest_paths:
            self._fg_noise_perturbers = {}
            for i in range(len(noise_manifest_paths)):
                if orig_sample_rate is None:
                    orig_sr = 16000
                else:
                    orig_sr = orig_sample_rate[i]
                self._fg_noise_perturbers[orig_sr] = NoisePerturbation(
                    manifest_path=noise_manifest_paths[i],
                    min_snr_db=min_snr_db[i],
                    max_snr_db=max_snr_db[i],
                    audio_tar_filepaths=noise_tar_filepaths[i],
                    orig_sr=orig_sr,
                )
        self._max_additions = max_additions
        self._max_duration = max_duration
        if bg_noise_manifest_paths:
            self._bg_noise_perturbers = {}
            for i in range(len(bg_noise_manifest_paths)):
                if bg_orig_sample_rate is None:
                    orig_sr = 16000
                else:
                    orig_sr = bg_orig_sample_rate[i]
                self._bg_noise_perturbers[orig_sr] = NoisePerturbation(
                    manifest_path=bg_noise_manifest_paths[i],
                    min_snr_db=bg_min_snr_db[i],
                    max_snr_db=bg_max_snr_db[i],
                    audio_tar_filepaths=bg_noise_tar_filepaths[i],
                    orig_sr=orig_sr,
                )

        self._apply_noise_rir = apply_noise_rir

    def perturb(self, data):
        prob = random.uniform(0.0, 1.0)

        if prob < self._rir_prob:
            self._rir_perturber.perturb(data)

        data_rms = data.rms_db

        if self._fg_noise_perturbers is not None and random.uniform(0.0, 1.0) < self._noise_prob:
            orig_sr = data.orig_sr
            if orig_sr not in self._fg_noise_perturbers:
                orig_sr = max(self._fg_noise_perturbers.keys())
            fg_perturber = self._fg_noise_perturbers[orig_sr]
            noise = fg_perturber.get_one_noise_sample(data.sample_rate)
            if self._apply_noise_rir:
                self._rir_perturber.perturb(noise)
            fg_perturber.perturb_with_foreground_noise(
                data, noise, data_rms=data_rms, max_noise_dur=self._max_duration, max_additions=self._max_additions
            )

        if self._bg_noise_perturbers is not None and random.uniform(0.0, 1.0) < self._bg_noise_prob:
            orig_sr = data.orig_sr
            if orig_sr not in self._bg_noise_perturbers:
                orig_sr = max(self._bg_noise_perturbers.keys())
            bg_perturber = self._bg_noise_perturbers[orig_sr]

            noise = bg_perturber.get_one_noise_sample(data.sample_rate)
            bg_perturber.perturb_with_input_noise(data, noise, data_rms=data_rms)


class TranscodePerturbation(Perturbation):
    """
    Audio codec augmentation. This implementation uses sox to transcode audio with low rate audio codecs,
    so users need to make sure that the installed sox version supports the codecs used here (G711 and amr-nb).

    Args:
        codecs (List[str]):A list of codecs to be trancoded to. Default is None.
        rng (int): Random seed. Default is None.
    """

    def __init__(self, codecs=None, rng=None):
        random.seed(rng) if rng else None
        self._codecs = codecs if codecs is not None else ["g711", "amr-nb", "ogg"]
        self.att_factor = 0.8  # to avoid saturation while writing to wav
        if codecs is not None:
            for codec in codecs:
                if codec not in ["g711", "amr-nb", "ogg"]:
                    raise ValueError(
                        f"TranscodePerturbation with {codec} isnot supported. Only {codecs} are supported"
                    )

    def perturb(self, data):
        max_level = np.max(np.abs(data._samples))
        if max_level > 0.8:
            norm_factor = self.att_factor / max_level
            norm_samples = norm_factor * data._samples
        else:
            norm_samples = data._samples
        orig_f = NamedTemporaryFile(suffix=".wav")
        sf.write(orig_f.name, norm_samples.transpose(), 16000)

        codec_ind = random.randint(0, len(self._codecs) - 1)
        if self._codecs[codec_ind] == "amr-nb":
            transcoded_f = NamedTemporaryFile(suffix="_amr.wav")
            rates = list(range(0, 4))
            rate = rates[random.randint(0, len(rates) - 1)]
            _ = subprocess.check_output(
                f"sox {orig_f.name} -V0 -C {rate} -t amr-nb - | sox -t amr-nb - -V0 -b 16 -r 16000 {transcoded_f.name}",
                shell=True,
            )
        elif self._codecs[codec_ind] == "ogg":
            transcoded_f = NamedTemporaryFile(suffix="_ogg.wav")
            rates = list(range(-1, 8))
            rate = rates[random.randint(0, len(rates) - 1)]
            _ = subprocess.check_output(
                f"sox {orig_f.name} -V0 -C {rate} -t ogg - | sox -t ogg - -V0 -b 16 -r 16000 {transcoded_f.name}",
                shell=True,
            )
        elif self._codecs[codec_ind] == "g711":
            transcoded_f = NamedTemporaryFile(suffix="_g711.wav")
            _ = subprocess.check_output(
                f"sox {orig_f.name} -V0  -r 8000 -c 1 -e a-law {transcoded_f.name} lowpass 3400 highpass 300",
                shell=True,
            )

        new_data = AudioSegment.from_file(transcoded_f.name, target_sr=16000)
        data._samples = new_data._samples[0 : data._samples.shape[0]]
        return


class RandomSegmentPerturbation(Perturbation):
    """
    Returns a random segment from input of duration "duration_sec".
    If duration_sec > input audio length, pad_to_duration determines the outcome.

    RandomSegmentPerturbation is intended for self-supervised learning.
    Not for supervised, as extracting corresponding text is not facilitated.


    Args:
        duration_sec (float): duration of the segment to be extracted
        pad_to_duration (bool): zero pad if length of input audio < duration_sec
        rng: Random seed. Default is None
        min_rms_db: Minimum RMS db value for the perturbed audio. Default is None
        max_trials: Maximum number of trials to find a segment with RMS db > min_rms_db. Default is 10
        verbose: If True, logs a warning if RMS db < min_rms_db after max_trials. Default is False
    """

    def __init__(
        self, duration_sec=32.0, pad_to_duration=False, rng=None, min_rms_db=None, max_trials=10, verbose=False
    ):
        if duration_sec <= 0:
            raise ValueError("duration_sec should be > 0")

        self._duration_sec = duration_sec
        self._pad_to_duration = pad_to_duration
        self._min_rms_db = min_rms_db
        self._max_trials = max_trials
        self._verbose = verbose
        random.seed(rng) if rng else None

    def perturb(self, data: AudioSegment):
        if self._duration_sec > data.duration:
            if not self._pad_to_duration:
                # don't do anything if pad_to_duration is False
                return
            start_time = 0.0
            pad_size = self._duration_sec * data.sample_rate - data.num_samples
            data.pad(pad_size=pad_size)
        else:
            start_time = random.uniform(0.0, data.duration - self._duration_sec)

        end_time = start_time + self._duration_sec
        new_data = copy.deepcopy(data)
        new_data.subsegment(start_time=start_time, end_time=end_time)
        if self._min_rms_db is not None:
            rms_db = new_data.rms_db if new_data.num_channels == 1 else min(new_data.rms_db)
            trial = 0
            while rms_db < self._min_rms_db and trial < self._max_trials:
                start_time = random.uniform(0.0, data.duration - self._duration_sec)
                end_time = start_time + self._duration_sec
                new_data = copy.deepcopy(data)
                new_data.subsegment(start_time=start_time, end_time=end_time)
                rms_db = new_data.rms_db if new_data.num_channels == 1 else min(new_data.rms_db)
                trial += 1
            if self._verbose and trial == self._max_trials and rms_db < self._min_rms_db:
                logging.warning(
                    f"Could not find a segment with RMS db > {self._min_rms_db} after {self._max_trials} trials."
                )
        data.subsegment(start_time=start_time, end_time=end_time)


perturbation_types = {
    "speed": SpeedPerturbation,
    "time_stretch": TimeStretchPerturbation,
    "gain": GainPerturbation,
    "silence": SilencePerturbation,
    "impulse": ImpulsePerturbation,
    "shift": ShiftPerturbation,
    "noise": NoisePerturbation,
    "noise_norm": NoisePerturbationWithNormalization,
    "white_noise": WhiteNoisePerturbation,
    "rir_noise_aug": RirAndNoisePerturbation,
    "transcode_aug": TranscodePerturbation,
    "random_segment": RandomSegmentPerturbation,
}


def register_perturbation(name: str, perturbation: Perturbation):
    if name in perturbation_types.keys():
        raise KeyError(
            f"Perturbation with the name {name} exists. " f"Type of perturbation : {perturbation_types[name]}."
        )

    perturbation_types[name] = perturbation


class AudioAugmentor(object):
    def __init__(self, perturbations=None, rng=None):
        random.seed(rng) if rng else None
        self._pipeline = perturbations if perturbations is not None else []

    def perturb(self, segment):
        for prob, p in self._pipeline:
            if random.random() < prob:
                p.perturb(segment)
        return

    def max_augmentation_length(self, length):
        newlen = length
        for prob, p in self._pipeline:
            newlen = p.max_augmentation_length(newlen)
        return newlen

    @classmethod
    def from_config(cls, config):
        ptbs = []
        for p in config:
            if p['aug_type'] not in perturbation_types:
                logging.warning("%s perturbation not known. Skipping.", p['aug_type'])
                continue
            perturbation = perturbation_types[p['aug_type']]
            ptbs.append((p['prob'], perturbation(**p['cfg'])))
        return cls(perturbations=ptbs)


def process_augmentations(augmenter, global_rank=0, world_size=1) -> Optional[AudioAugmentor]:
    """Process list of online data augmentations.
    Accepts either an AudioAugmentor object with pre-defined augmentations,
    or a dictionary that points to augmentations that have been defined.
    If a dictionary is passed, must follow the below structure:
    Dict[str, Dict[str, Any]]: Which refers to a dictionary of string
    names for augmentations, defined in `asr/parts/perturb.py`.
    The inner dictionary may contain key-value arguments of the specific
    augmentation, along with an essential key `prob`. `prob` declares the
    probability of the augmentation being applied, and must be a float
    value in the range [0, 1].
    # Example in YAML config file
    Augmentations are generally applied only during training, so we can add
    these augmentations to our yaml config file, and modify the behaviour
    for training and evaluation.
    ```yaml
    AudioToSpeechLabelDataLayer:
        ...  # Parameters shared between train and evaluation time
        train:
            augmentor:
                shift:
                    prob: 0.5
                    min_shift_ms: -5.0
                    max_shift_ms: 5.0
                white_noise:
                    prob: 1.0
                    min_level: -90
                    max_level: -46
                ...
        eval:
            ...
    ```
    Then in the training script,
    ```python
    import copy
    from ruamel.yaml import YAML
    yaml = YAML(typ="safe")
    with open(model_config) as f:
        params = yaml.load(f)
    # Train Config for Data Loader
    train_dl_params = copy.deepcopy(params["AudioToTextDataLayer"])
    train_dl_params.update(params["AudioToTextDataLayer"]["train"])
    del train_dl_params["train"]
    del train_dl_params["eval"]
    data_layer_train = nemo_asr.AudioToTextDataLayer(
        ...,
        **train_dl_params,
    )
    # Evaluation Config for Data Loader
    eval_dl_params = copy.deepcopy(params["AudioToTextDataLayer"])
    eval_dl_params.update(params["AudioToTextDataLayer"]["eval"])
    del eval_dl_params["train"]
    del eval_dl_params["eval"]
    data_layer_eval = nemo_asr.AudioToTextDataLayer(
        ...,
        **eval_dl_params,
    )
    ```
    # Registering your own Augmentations
    To register custom augmentations to obtain the above convenience of
    the declaring the augmentations in YAML, you can put additional keys in
    `perturbation_types` dictionary as follows.
    ```python
    from nemo.collections.asr.parts import perturb
    # Define your own perturbation here
    class CustomPerturbation(perturb.Perturbation):
        ...
    perturb.register_perturbation(name_of_perturbation, CustomPerturbation)
    ```
    Args:
        augmenter: AudioAugmentor object or
            dictionary of str -> kwargs (dict) which is parsed and used
            to initialize an AudioAugmentor.
            Note: It is crucial that each individual augmentation has
            a keyword `prob`, that defines a float probability in the
            the range [0, 1] of this augmentation being applied.
            If this keyword is not present, then the augmentation is
            disabled and a warning is logged.
    Returns: AudioAugmentor object
    """
    if augmenter is None:
        return None

    if isinstance(augmenter, AudioAugmentor):
        return augmenter

    augmenter_types = {dict}
    if HAVE_OMEGACONG_WEBDATASET:
        augmenter_types = {dict, DictConfig}
    if not type(augmenter) in augmenter_types:
        raise ValueError("Cannot parse augmenter. Must be a dict or an AudioAugmentor object ")

    if HAVE_OMEGACONG_WEBDATASET and isinstance(augmenter, DictConfig):
        augmenter = OmegaConf.to_container(augmenter, resolve=True)

    augmenter = copy.deepcopy(augmenter)

    augmentations = []
    for augment_name, augment_kwargs in augmenter.items():
        prob = augment_kwargs.get('prob', None)

        if prob is None:
            raise KeyError(
                f'Augmentation "{augment_name}" will not be applied as '
                f'keyword argument "prob" was not defined for this augmentation.'
            )

        else:
            _ = augment_kwargs.pop('prob')

            if prob < 0.0 or prob > 1.0:
                raise ValueError("`prob` must be a float value between 0 and 1.")

            try:
                augmentation_class = perturbation_types[augment_name]
                if 'global_rank' in inspect.signature(augmentation_class).parameters:
                    augment_kwargs['global_rank'] = global_rank
                if 'world_size' in inspect.signature(augmentation_class).parameters:
                    augment_kwargs['world_size'] = world_size
                augmentation = augmentation_class(**augment_kwargs)
                augmentations.append([prob, augmentation])
            except KeyError:
                raise KeyError(f"Invalid perturbation name. Allowed values : {perturbation_types.keys()}")

    augmenter = AudioAugmentor(perturbations=augmentations)
    return augmenter


class AugmentationDataset(IterableDataset):
    """
    A class that loads tarred audio files and cycles over the files in the dataset.
    Accepts a single comma-separated JSON manifest file (in the same style as for the AudioToCharDataset/AudioToBPEDataset),
    as well as the path(s) to the tarball(s) containing the wav files. Each line of the manifest should
    contain the information for one audio file, including at least the transcript and name of the audio
    file within the tarball.
    Valid formats for the audio_tar_filepaths argument include:
    (1) a single string that can be brace-expanded, e.g. 'path/to/audio.tar' or 'path/to/audio_{1..100}.tar.gz', or
    (2) a list of file paths that will not be brace-expanded, e.g. ['audio_1.tar', 'audio_2.tar', ...].
    Note: For brace expansion in (1), there may be cases where `{x..y}` syntax cannot be used due to shell interference.
    This occurs most commonly inside SLURM scripts. Therefore we provide a few equivalent replacements.
    Supported opening braces - { <=> (, [, < and the special tag _OP_.
    Supported closing braces - } <=> ), ], > and the special tag _CL_.
    For SLURM based tasks, we suggest the use of the special tags for ease of use.
    See the WebDataset documentation for more information about accepted data and input formats.
    """

    def __init__(
        self,
        manifest_path: str,
        tar_filepaths: Union[str, List[str]],
        shuffle_n: int = 128,
        rank: int = 0,
        world_size: int = 1,
        shard_strategy: str = "replicate",
    ):
        # import here to avoid circular import error
        from nemo.collections.asr.data.audio_to_text import expand_sharded_filepaths

        self._manifest = collections.ASRAudioText(manifest_path, parser=parsers.make_parser([]), index_by_file_id=True)

        tar_filepaths = expand_sharded_filepaths(
            tar_filepaths, shard_strategy=shard_strategy, world_size=world_size, global_rank=rank
        )

        if not HAVE_OMEGACONG_WEBDATASET:
            raise LightningNotInstalledException(self)
        self.audio_dataset = wds.DataPipeline(
            wds.SimpleShardList(urls=tar_filepaths),
            wds.shuffle(shuffle_n),
            wds.tarfile_to_samples(),
            wds.rename(audio='wav;ogg;flac', key='__key__'),
            wds.to_tuple('audio', 'key'),
            self._loop_offsets,
        )

    def __len__(self):
        return len(self._manifest)

    def _loop_offsets(self, iterator):
        """This function is used to iterate through utterances with different offsets for each file."""

        class TarredAudioLoopOffsets:
            def __init__(self, collection):
                self.iterator = iterator
                self.collection = collection
                self.current_fn = None
                self.current_bytes = None
                self.offset_id = 0

            def __iter__(self):
                return self

            def __next__(self):
                if self.current_fn is None:
                    self.current_bytes, self.current_fn = next(self.iterator)
                    self.offset_id = 0
                else:
                    offset_list = self.collection.mapping[self.current_fn]
                    if len(offset_list) == self.offset_id + 1:
                        self.current_bytes, self.current_fn = next(self.iterator)
                        self.offset_id = 0
                    else:
                        self.offset_id += 1

                return self.current_bytes, self.current_fn, self.offset_id

        return TarredAudioLoopOffsets(self._manifest)

    def __iter__(self):
        audio_iter = iter(self.audio_dataset)

        while True:
            try:
                audio_bytes, audio_filename, offset_id = next(audio_iter)
                file_id, _ = os.path.splitext(os.path.basename(audio_filename))
                manifest_idx = self._manifest.mapping[file_id][offset_id]
                manifest_entry = self._manifest[manifest_idx]

                # Convert audio bytes to IO stream for processing (for SoundFile to read)
                audio_file = io.BytesIO(audio_bytes)
                yield audio_file, file_id, manifest_entry
            except StopIteration:
                audio_iter = iter(self.audio_dataset)



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Type

import numpy as np
import torch

from nemo.collections.common.parts.utils import mask_sequence_tensor
from nemo.core.classes import NeuralModule, typecheck
from nemo.core.neural_types import FloatType, LengthsType, NeuralType, SpectrogramType, VoidType
from nemo.utils import logging


class StochasticDifferentialEquation(NeuralModule, ABC):
    """Base class for stochastic differential equations."""

    def __init__(self, time_min: float, time_max: float, num_steps: int):
        super().__init__()

        # min and max time
        if time_min <= 0:
            raise ValueError(f'time_min should be positive, current value {time_min}')

        if time_max <= time_min:
            raise ValueError(f'time_max should be larger than time_min, current max {time_max} and min {time_min}')

        self.time_min = time_min
        self.time_max = time_max

        # number of steps
        if num_steps <= 0:
            raise ValueError(f'num_steps needs to be positive: current value {num_steps}')

        self.num_steps = num_steps

    @property
    def dt(self) -> float:
        """Time step for this SDE.
        This denotes the step size between `0` and `self.time_max` when using `self.num_steps`.
        """
        return self.time_max / self.num_steps

    @property
    def time_delta(self) -> float:
        """Time range for this SDE."""
        return self.time_max - self.time_min

    def generate_time(self, size: int, device: torch.device) -> torch.Tensor:
        """Generate random time steps in the valid range.

        Time steps are generated between `self.time_min` and `self.time_max`.

        Args:
            size: number of samples
            device: device to use

        Returns:
            A tensor of floats with shape (size,)
        """
        time = torch.rand(size, device=device) * self.time_delta + self.time_min
        return time

    @abstractmethod
    def coefficients(self, state: torch.Tensor, time: torch.Tensor, **kwargs) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            state: tensor of shape (B, C, D, T)
            time: tensor of shape (B,)

        Returns:
            Tuple with drift and diffusion coefficients.
        """
        pass

    @typecheck(
        input_types={
            "prior_mean": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
        },
        output_types={
            "sample": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
        },
    )
    @abstractmethod
    def prior_sampling(self, prior_mean: torch.Tensor) -> torch.Tensor:
        """Generate a sample from the prior distribution p_T.

        Args:
            prior_mean: Mean of the prior distribution

        Returns:
            A sample from the prior distribution.
        """
        pass

    def discretize(
        self, *, state: torch.Tensor, time: torch.Tensor, state_length: Optional[torch.Tensor] = None, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Assume we have the following SDE:

            dx = drift(x, t) * dt + diffusion(x, t) * dwt

        where `wt` is the standard Wiener process.

        We assume the following discretization:

            new_state = current_state + total_drift + total_diffusion * z_norm

        where `z_norm` is sampled from normal distribution with zero mean and unit variance.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            state_length: length of the valid time steps for each example in the batch, shape (B,)
            **kwargs: other parameters

        Returns:
            Drift and diffusion.
        """
        # Get coefficients
        drift_coefficient, diffusion_coefficient = self.coefficients(
            state=state, time=time, state_length=state_length, **kwargs
        )

        # Discretized drift
        drift = drift_coefficient * self.dt

        # Note:
        # Scale with sqrt(dt) because z_norm is sampled from a normal distribution with zero mean and
        # unit variance and dwt is normally distributed with zero mean and variance dt
        diffusion = diffusion_coefficient * np.sqrt(self.dt)

        return drift, diffusion

    @abstractmethod
    def copy(self):
        """Create a copy of this SDE."""
        pass

    def __repr__(self):
        desc = f'{self.__class__.__name__}(time_min={self.time_min}, time_max={self.time_max}, num_steps={self.num_steps})'
        desc += f'\n\tdt:         {self.dt}'
        desc += f'\n\ttime_delta: {self.time_delta}'
        return desc


class OrnsteinUhlenbeckVarianceExplodingSDE(StochasticDifferentialEquation):
    """This class implements the Ornstein-Uhlenbeck SDE with variance exploding noise schedule.

    The SDE is given by:

        dx = theta * (y - x) dt + g(t) dw

    where `theta` is the stiffness parameter and `g(t)` is the diffusion coefficient:

        g(t) = std_min * (std_max/std_min)^t * sqrt(2 * log(std_max/std_min))

    References:
        Richter et al., Speech Enhancement and Dereverberation with Diffusion-based Generative Models, Tr. ASLP 2023
    """

    def __init__(
        self,
        stiffness: float,
        std_min: float,
        std_max: float,
        num_steps: int = 100,
        time_min: float = 3e-2,
        time_max: float = 1.0,
        eps: float = 1e-8,
    ):
        super().__init__(time_min=time_min, time_max=time_max, num_steps=num_steps)

        # Small regularization
        if eps <= 0:
            raise ValueError(f'eps should be positive, current value {eps}')
        self.eps = eps

        # stifness
        self.stiffness = stiffness

        # noise schedule
        if std_min <= 0:
            raise ValueError(f'std_min should be positive, current value {std_min}')

        if std_max <= std_min:
            raise ValueError(f'std_max should be larger than std_min, current max {std_max} and min {std_min}')

        self.std_min = std_min
        self.std_max = std_max

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tstiffness:     %s', self.stiffness)
        logging.debug('\tstd_min:       %s', self.std_min)
        logging.debug('\tstd_max:       %s', self.std_max)
        logging.debug('\tnum_steps:     %s', self.num_steps)
        logging.debug('\ttime_min:      %s', self.time_min)
        logging.debug('\ttime_max:      %s', self.time_max)
        logging.debug('\teps:           %s', self.eps)

    @property
    def std_ratio(self) -> float:
        return self.std_max / (self.std_min + self.eps)

    @property
    def log_std_ratio(self) -> float:
        return np.log(self.std_ratio + self.eps)

    @typecheck(
        input_types={
            "state": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "prior_mean": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "time": NeuralType(tuple('B'), FloatType()),
        },
        output_types={
            "mean": NeuralType(('B', 'C', 'D', 'T'), FloatType()),
        },
    )
    def perturb_kernel_mean(self, state: torch.Tensor, prior_mean: torch.Tensor, time: torch.Tensor) -> torch.Tensor:
        """Return the mean of the perturbation kernel for this SDE.

        Args:
            state: current state of the process, shape (B, C, D, T)
            prior_mean: mean of the prior distribution
            time: current time of the process, shape (B,)

        Returns:
            A tensor of shape (B, C, D, T)
        """
        # exponential weighting
        weight = torch.exp(-self.stiffness * time)

        # view as [B, C, D, T]
        weight = weight.view(-1, 1, 1, 1)

        # closed-form mean
        mean = weight * state + (1 - weight) * prior_mean

        return mean

    @typecheck(
        input_types={
            "time": NeuralType(tuple('B'), FloatType()),
        },
        output_types={
            "std": NeuralType(tuple('B'), FloatType()),
        },
    )
    def perturb_kernel_std(self, time: torch.Tensor) -> torch.Tensor:
        """Return the standard deviation of the perturbation kernel for this SDE.

        Note that the standard deviation depends on the time and the noise schedule,
        which is parametrized using `self.stiffness`, `self.std_min` and `self.std_max`.

        Args:
            time: current time of the process, shape (B,)

        Returns:
            A tensor of shape (B,)
        """
        var = (self.std_min**2) * self.log_std_ratio
        var *= torch.pow(self.std_ratio, 2 * time) - torch.exp(-2 * self.stiffness * time)
        var /= self.stiffness + self.log_std_ratio
        std = torch.sqrt(var)
        return std

    @typecheck(
        input_types={
            "state": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "prior_mean": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "time": NeuralType(tuple('B'), FloatType()),
        },
        output_types={
            "mean": NeuralType(('B', 'C', 'D', 'T'), FloatType()),
            "std": NeuralType(('B', 'C', 'D', 'T'), FloatType()),
        },
    )
    def perturb_kernel_params(self, state: torch.Tensor, prior_mean: torch.Tensor, time: torch.Tensor) -> torch.Tensor:
        """Return the mean and standard deviation of the perturbation kernel for this SDE.

        Args:
            state: current state of the process, shape (B, C, D, T)
            prior_mean: mean of the prior distribution
            time: current time of the process, shape (B,)
        """
        assert torch.all(time <= self.time_max)
        assert torch.all(time >= self.time_min)

        # compute the mean
        mean = self.perturb_kernel_mean(state=state, prior_mean=prior_mean, time=time)

        # compute the standard deviation
        std = self.perturb_kernel_std(time=time)
        # view as [B, C, D, T]
        std = std.view(-1, 1, 1, 1)

        return mean, std

    @typecheck(
        input_types={
            "state": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "time": NeuralType(tuple('B'), VoidType()),
            "prior_mean": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "state_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        },
        output_types={
            "drift_coefficient": NeuralType(('B', 'C', 'D', 'T'), FloatType()),
            "diffusion_coefficient": NeuralType(('B', 'C', 'D', 'T'), FloatType()),
        },
    )
    def coefficients(
        self,
        state: torch.Tensor,
        time: torch.Tensor,
        prior_mean: torch.Tensor,
        state_length: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute drift and diffusion coefficients for this SDE.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            prior_mean: mean of the prior distribution
            state_length: length of the valid time steps for each example in the batch

        Returns:
            Drift and diffusion coefficients.
        """
        # Drift coefficient
        drift_coefficient = self.stiffness * (prior_mean - state)

        # Diffusion coefficient
        diffusion_coefficient = self.std_min * torch.pow(self.std_ratio, time) * np.sqrt(2 * self.log_std_ratio)
        # View in the same shape as the state
        diffusion_coefficient = diffusion_coefficient.view(-1, *([1] * (state.dim() - 1)))

        if state_length is not None:
            drift_coefficient = mask_sequence_tensor(drift_coefficient, state_length)
            diffusion_coefficient = mask_sequence_tensor(diffusion_coefficient, state_length)

        return drift_coefficient, diffusion_coefficient

    def prior_sampling(self, prior_mean: torch.Tensor) -> torch.Tensor:
        """Generate a sample from the prior distribution p_T.

        Args:
            prior_mean: Mean of the prior distribution
        """
        # Final time step for all samples in the batch
        time = self.time_max * torch.ones(prior_mean.shape[0], device=prior_mean.device)

        # Compute the std of the prior distribution
        std = self.perturb_kernel_std(time=time)

        # view as [B, C, D, T]
        std = std.view(-1, 1, 1, 1)

        # Generate a sample from a normal distribution centered at prior_mean
        sample = prior_mean + torch.randn_like(prior_mean) * std

        return sample

    def copy(self):
        return OrnsteinUhlenbeckVarianceExplodingSDE(
            stiffness=self.stiffness,
            std_min=self.std_min,
            std_max=self.std_max,
            num_steps=self.num_steps,
            time_min=self.time_min,
            time_max=self.time_max,
            eps=self.eps,
        )

    def __repr__(self):
        desc = f'{self.__class__.__name__}(stiffness={self.stiffness}, std_min={self.std_min}, std_max={self.std_max}, num_steps={self.num_steps}, time_min={self.time_min}, time_max={self.time_max}, eps={self.eps})'
        desc += f'\n\tdt:         {self.dt}'
        desc += f'\n\ttime_delta: {self.time_delta}'
        desc += f'\n\tstd_ratio:  {self.std_ratio}'
        desc += f'\n\tlog_std_ratio:  {self.log_std_ratio}'

        return desc


class ReverseStochasticDifferentialEquation(StochasticDifferentialEquation):
    def __init__(self, *, sde: Type[StochasticDifferentialEquation], score_estimator: Type[NeuralModule]):
        """Use the forward SDE and a score estimator to define the reverse SDE.

        Args:
            sde: forward SDE
            score_estimator: neural score estimator
        """
        super().__init__(time_min=sde.time_min, time_max=sde.time_max, num_steps=sde.num_steps)
        self.score_estimator = score_estimator
        self.forward_sde = sde

        logging.debug('Initialized %s', self.__class__.__name__)

    def coefficients(
        self,
        state: torch.Tensor,
        time: torch.Tensor,
        score_condition: Optional[torch.Tensor] = None,
        state_length: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute drift and diffusion coefficients for the reverse SDE.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
        """
        raise NotImplementedError('Coefficients not necessary for the reverse SDE.')

    def prior_sampling(self, shape: torch.Size, device: torch.device) -> torch.Tensor:
        """Prior sampling is not necessary for the reverse SDE."""
        raise NotImplementedError('Prior sampling not necessary for the reverse SDE.')

    def discretize(
        self,
        *,
        state: torch.Tensor,
        time: torch.Tensor,
        score_condition: Optional[torch.Tensor] = None,
        state_length: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Discretize the reverse SDE.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            score_condition: condition for the score estimator
            state_length: length of the valid time steps for each example in the batch
            **kwargs: other parameters for discretization of the forward SDE
        """
        # Drift and diffusion from the forward SDE
        forward_drift, forward_diffusion = self.forward_sde.discretize(state=state, time=time, **kwargs)

        # For input for the score estimator:
        # - if no condition is provided, use the state
        # - if a condition is provided, concatenate the state and the condition along the channel dimension
        score_input = state if score_condition is None else torch.cat([state, score_condition], dim=1)

        # Estimate score
        score, _ = self.score_estimator(input=score_input, input_length=state_length, condition=time)

        # Adjust drift
        drift = forward_drift - forward_diffusion.pow(2) * score

        # Adjust diffusion
        diffusion = forward_diffusion

        if state_length is not None:
            drift = mask_sequence_tensor(drift, state_length)
            diffusion = mask_sequence_tensor(diffusion, state_length)

        return drift, diffusion

    def copy(self):
        return ReverseStochasticDifferentialEquation(sde=self.forward_sde.copy(), score_estimator=self.score_estimator)

    def __repr__(self):
        desc = f'{self.__class__.__name__}(sde={self.forward_sde}, score_estimator={self.score_estimator})'
        return desc


class PredictorCorrectorSampler(NeuralModule):
    """Predictor-Corrector sampler for the reverse SDE.

    Args:
        sde: forward SDE
        score_estimator: neural score estimator
        predictor: predictor for the reverse process
        corrector: corrector for the reverse process
        num_steps: number of time steps for the reverse process
        num_corrector_steps: number of corrector steps
        time_max: maximum time
        time_min: minimum time
        snr: SNR for Annealed Langevin Dynamics
        output_type: type of the output ('state' for the final state, or 'mean' for the mean of the final state)

    References:
        - Song et al., Score-based generative modeling through stochastic differential equations, 2021
    """

    def __init__(
        self,
        sde,
        score_estimator,
        predictor: str = 'reverse_diffusion',
        corrector: str = 'annealed_langevin_dynamics',
        num_steps: int = 50,
        num_corrector_steps: int = 1,
        time_max: Optional[float] = None,
        time_min: Optional[float] = None,
        snr: float = 0.5,
        output_type: str = 'mean',
    ):
        super().__init__()
        # Create a copy of SDE
        self.sde = sde.copy()

        # Update SDE parameters for sampling
        if time_max is not None:
            self.sde.time_max = time_max
            logging.info('sde.time_max set to: %s', self.sde.time_max)

        if time_min is not None:
            self.sde.time_min = time_min
            logging.info('sde.time_min set to: %s', self.sde.time_min)

        self.sde.num_steps = num_steps
        logging.info('sde.num_steps set to: %s', self.sde.num_steps)

        # Update local values
        self.time_max = self.sde.time_max
        self.time_min = self.sde.time_min
        self.num_steps = self.sde.num_steps

        # Predictor setup
        if predictor == 'reverse_diffusion':
            self.predictor = ReverseDiffusionPredictor(sde=self.sde, score_estimator=score_estimator)
        else:
            raise RuntimeError(f'Unexpected predictor: {predictor}')

        # Corrector setup
        if corrector == 'annealed_langevin_dynamics':
            self.corrector = AnnealedLangevinDynamics(
                sde=self.sde, score_estimator=score_estimator, snr=snr, num_steps=num_corrector_steps
            )
        else:
            raise RuntimeError(f'Unexpected corrector: {corrector}')

        if output_type not in ['mean', 'state']:
            raise ValueError(f'Unexpected output type: {output_type}')
        self.output_type = output_type

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tpredictor:           %s', predictor)
        logging.debug('\tcorrector:           %s', corrector)
        logging.debug('\tnum_steps:           %s', self.num_steps)
        logging.debug('\ttime_min:            %s', self.time_min)
        logging.debug('\ttime_max:            %s', self.time_max)
        logging.debug('\tnum_corrector_steps: %s', num_corrector_steps)
        logging.debug('\tsnr:                 %s', snr)
        logging.debug('\toutput_type:         %s', self.output_type)

    @typecheck(
        input_types={
            "prior_mean": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "score_condition": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType(), optional=True),
            "state_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        },
        output_types={
            "sample": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "state_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        },
    )
    @torch.inference_mode()
    def forward(
        self, prior_mean: torch.Tensor, score_condition: torch.Tensor, state_length: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Takes prior (noisy) mean and generates a sample by solving the reverse SDE.

        Args:
            prior_mean: mean for the prior distribution, e.g., noisy observation
            score_condition: conditioning for the score estimator
            state_length: length of the valid time steps for each example in the batch

        Returns:
            Generated `sample` and the corresponding `sample_length`.
        """
        # Sample from the prior distribution
        state = self.sde.prior_sampling(prior_mean=prior_mean)

        if state_length is not None:
            state = mask_sequence_tensor(state, state_length)

        # Time steps for evaluation
        time_steps = torch.linspace(self.time_max, self.time_min, self.num_steps, device=state.device)

        # Sampling
        for t in time_steps:
            # time steps for the whole batch
            time = t * torch.ones(state.shape[0], device=state.device)

            # corrector step
            state, _ = self.corrector(
                state=state, time=time, score_condition=score_condition, state_length=state_length
            )

            # predictor step
            state, state_mean = self.predictor(
                state=state,
                time=time,
                score_condition=score_condition,
                prior_mean=prior_mean,
                state_length=state_length,
            )

        # Final output
        if self.output_type == 'state':
            sample = state
        elif self.output_type == 'mean':
            sample = state_mean
        else:
            raise RuntimeError(f'Unexpected output type: {self.output_type}')

        if state_length is not None:
            sample = mask_sequence_tensor(sample, state_length)

        return sample, state_length


class Predictor(torch.nn.Module, ABC):
    """Predictor for the reverse process.

    Args:
        sde: forward SDE
        score_estimator: neural score estimator
    """

    def __init__(self, sde, score_estimator):
        super().__init__()
        self.reverse_sde = ReverseStochasticDifferentialEquation(sde=sde, score_estimator=score_estimator)

    @abstractmethod
    @torch.inference_mode()
    def forward(
        self,
        *,
        state: torch.Tensor,
        time: torch.Tensor,
        score_condition: Optional[torch.Tensor] = None,
        state_length: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        """Predict the next state of the reverse process.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            score_condition: conditioning for the score estimator
            state_length: length of the valid time steps for each example in the batch

        Returns:
            New state and mean.
        """
        pass


class ReverseDiffusionPredictor(Predictor):
    """Predict the next state of the reverse process using the reverse diffusion process.

    Args:
        sde: forward SDE
        score_estimator: neural score estimator
    """

    def __init__(self, sde, score_estimator):
        super().__init__(sde=sde, score_estimator=score_estimator)

    @torch.inference_mode()
    def forward(self, *, state, time, score_condition=None, state_length=None, **kwargs):
        """Predict the next state of the reverse process using the reverse diffusion process.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            score_condition: conditioning for the score estimator
            state_length: length of the valid time steps for each example in the batch

        Returns:
            New state and mean of the diffusion process.
        """
        drift, diffusion = self.reverse_sde.discretize(
            state=state, time=time, score_condition=score_condition, state_length=state_length, **kwargs
        )

        # Generate a random sample from a standard normal distribution
        z_norm = torch.randn_like(state)

        # Compute the mean of the next state
        mean = state - drift

        # Compute new state by sampling
        new_state = mean + diffusion * z_norm

        if state_length is not None:
            new_state = mask_sequence_tensor(new_state, state_length)
            mean = mask_sequence_tensor(mean, state_length)

        return new_state, mean


class Corrector(NeuralModule, ABC):
    """Corrector for the reverse process.

    Args:
        sde: forward SDE
        score_estimator: neural score estimator
        snr: SNR for Annealed Langevin Dynamics
        num_steps: number of steps for the corrector
    """

    def __init__(
        self,
        sde: Type[StochasticDifferentialEquation],
        score_estimator: Type[NeuralModule],
        snr: float,
        num_steps: int,
    ):
        super().__init__()
        self.sde = sde
        self.score_estimator = score_estimator
        self.snr = snr
        self.num_steps = num_steps

        logging.debug('Initialized %s with', self.__class__.__name__)
        logging.debug('\tsnr:             %s', snr)
        logging.debug('\tnum_steps:       %s', num_steps)

    @abstractmethod
    @typecheck(
        input_types={
            "state": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
            "time": NeuralType(tuple('B'), FloatType()),
            "score_condition": NeuralType(('B', 'C', 'D', 'T'), VoidType(), optional=True),
            "state_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        },
        output_types={
            "state": NeuralType(('B', 'C', 'D', 'T'), VoidType()),
        },
    )
    @torch.inference_mode()
    def forward(self, state, time, score_condition=None, state_length=None):
        """
        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            score_condition: conditioning for the score estimator
            state_length: length of the valid time steps for each example in the batch

        Returns:
            New state and mean.
        """
        pass


class AnnealedLangevinDynamics(Corrector):
    """Annealed Langevin Dynamics for the reverse process.

    References:
        - Song et al., Score-based generative modeling through stochastic differential equations, 2021
    """

    def __init__(self, sde, **kwargs):
        if not isinstance(sde, OrnsteinUhlenbeckVarianceExplodingSDE):
            raise ValueError(f'Expected an instance of OrnsteinUhlenbeckVarianceExplodingSDE, got {type(sde)}')
        super().__init__(sde=sde, **kwargs)

    @torch.inference_mode()
    def forward(self, state, time, score_condition=None, state_length=None):
        """Correct the state using Annealed Langevin Dynamics.

        Args:
            state: current state of the process, shape (B, C, D, T)
            time: current time of the process, shape (B,)
            score_condition: conditioning for the score estimator
            state_length: length of the valid time steps for each example in the batch

        Returns:
            New state and mean of the diffusion process.

        References:
            Alg. 4 in http://arxiv.org/abs/2011.13456
        """
        # Compute the standard deviation of the diffusion process
        std = self.sde.perturb_kernel_std(time=time)
        # View as [B, 1, 1, 1]
        std = std.view(-1, *([1] * (state.dim() - 1)))

        for i in range(self.num_steps):
            # prepare input for the score estimator, concatenate conditioning along the channel dimension
            score_input = state if score_condition is None else torch.cat([state, score_condition], dim=1)

            # calculate the score
            score, _ = self.score_estimator(input=score_input, input_length=state_length, condition=time)

            # generate a sample from a standard normal distribution
            z_norm = torch.randn_like(state)

            # compute the step size
            # note: this is slightly different than in the paper, where std = ||z_norm||_2 / ||score||_2
            step_size = 2 * (self.snr * std).pow(2)

            # update the mean
            mean = state + step_size * score

            # update the state
            state = mean + z_norm * torch.sqrt(step_size * 2)

        if state_length is not None:
            state = mask_sequence_tensor(state, state_length)
            mean = mask_sequence_tensor(mean, state_length)

        return state, mean



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
# pylint: skip-file

import copy
import math
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Union

import lightning.pytorch as pl
import torch
from megatron.core import parallel_state
from megatron.core.num_microbatches_calculator import update_num_microbatches
from omegaconf.omegaconf import DictConfig, OmegaConf
from pytorch_lightning.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from torch.utils.data import DataLoader

from nemo.collections.asr.parts.preprocessing.perturb import process_augmentations
from nemo.collections.common.data.blendable_dataset import BlendableDataset
from nemo.collections.common.data.dataset import ConcatMapDataset
from nemo.collections.common.data.lhotse import get_lhotse_dataloader_from_config
from nemo.collections.common.tokenizers import TokenizerSpec
from nemo.collections.multimodal.speech_llm.parts.utils.data_utils import get_text_processor_from_cfg
from nemo.collections.speechlm.data.data_sampler import SpeechLMDataSampler
from nemo.collections.speechlm.data.dataset.audio_text_dataset import (
    get_audio_text_dataset_from_config,
    get_tarred_audio_text_dataset_from_config,
)
from nemo.collections.speechlm.data.dataset.audio_text_lhotse_dataset import MultimodalConversationDataset
from nemo.collections.speechlm.data.text_processing import MultimodalConversationTextProcessor
from nemo.lightning.data import WrappedDataLoader
from nemo.lightning.io.mixin import IOMixin
from nemo.utils import logging


class AudioToTextDataModule(pl.LightningDataModule, IOMixin):
    """
    Data module for speech-to-text LLM.
    """

    def __init__(self, config: Union[DictConfig, Dict], tokenizer: TokenizerSpec):
        super().__init__()
        self.cfg = OmegaConf.create(config) if not isinstance(config, DictConfig) else config
        self.tokenizer = tokenizer
        self._train_ds = None
        self._validation_ds = None
        self._test_ds = None
        self._validation_names = None
        self._test_names = None
        self.init_global_step = 0
        self.text_processor = None
        self._num_validation_dl = None
        self._num_test_dl = None

    @property
    def global_batch_size(self):
        """
        get the global batch size
        """
        return self.data_cfg.global_batch_size

    @property
    def micro_batch_size(self):
        """
        get the micro batch size
        """
        return self.data_cfg.micro_batch_size

    @property
    def seq_length(self):
        """
        get the max sequence length
        """
        return self.data_cfg.max_seq_length

    @property
    def data_cfg(self):
        """
        get the common data configuration
        """
        if 'common' not in self.cfg:
            raise ValueError("`common` configuration is missing in the data config")
        return self.cfg.common

    def get_text_processor(self):
        """
        Builds the text processor for SpeechLM
        """
        data_cfg = self.data_cfg

        if data_cfg.get('prompt_format', None):
            prompt_format = data_cfg['prompt_format']
        else:
            logging.warning(f"Prompt format is not specified in the data config. Using default prompt format `plain`.")
            prompt_format = "plain"

        # Text processor for lhotse datasets
        text_processor = MultimodalConversationTextProcessor(
            self.tokenizer,
            prompt_format=prompt_format,
            max_seq_length=data_cfg["max_seq_length"],
            add_boa_eoa=data_cfg.get("add_boa_eoa", False),
            boa_string=data_cfg.get("boa_string", "<BOA>"),
            eoa_string=data_cfg.get("eoa_string", "<EOA>"),
        )

        return text_processor

    def prepare_data(self):
        """
        download, IO, etc. Useful with shared filesystems
        only called on 1 GPU/TPU in distributed
        """
        pass

    def setup(self, stage=None):
        """
        build datasets, text processor and data sampler
        """
        self.text_processor = self.get_text_processor()

        # make assignments here (train/val/test split)
        # called on every process in DDP
        if stage == 'fit' or stage is None:
            self._train_ds = self._create_dataset('train')
            self._validation_ds = self._create_dataset('validation')
        elif stage == 'validate' or stage is None:
            self._validation_ds = self._create_dataset('validation')
        if stage == 'test' or stage is None:
            self._test_ds = self._create_dataset('test')

        self.data_sampler = SpeechLMDataSampler(
            seq_len=self.seq_length,
            micro_batch_size=self.micro_batch_size,
            global_batch_size=self.global_batch_size,
            rampup_batch_size=self.data_cfg.get("rampup_batch_size", None),
            dataloader_type="batch",  # "batch" should be used for SFT,
        )

        # Follows the calculation in `nemo.collections.nlp.data.language_modeling.megatron.
        # base_dataset_utils.get_datasets_weights_and_num_samples`
        self.max_train_samples = int(math.ceil(self.global_batch_size * self.trainer.max_steps * 1.005))

    @lru_cache
    def _create_dataset(self, mode: str):
        """
        Create the datasets for each of train/validation/test/predict mode
        """
        data_cfg = self.cfg.get(f"{mode}_ds", None)
        if data_cfg is None:
            logging.info(f"Skipping {mode} dataset creation as it is not specified in the config: {self.cfg}")
            return None

        if 'augmentor' in data_cfg:
            augmentor = process_augmentations(
                data_cfg['augmentor'],
                global_rank=parallel_state.get_data_parallel_rank(),
                world_size=parallel_state.get_data_parallel_world_size(),
            )
        else:
            augmentor = None

        # Notably, the data weights are controlled by either bucketing_weights
        # or concat_sampling_probabilities depending on the dataset type.
        if data_cfg.get("use_lhotse"):
            logging.info(f"Creating Lhotse dataset for {mode}")
            if mode != 'train':
                setattr(self, f"_{mode}_names", self._parse_lhotse_data_name(mode))

            return MultimodalConversationDataset(
                text_processor=self.text_processor,
                default_context=data_cfg.get('default_context', 'listen to the audio'),
                tokens_to_generate=data_cfg.get('tokens_to_generate', 0),
                pad_to_max_length=data_cfg.get('pad_to_max_length', False),
                max_seq_length=data_cfg["max_seq_length"],
                context_key=data_cfg.get('context_key'),
                answer_key=data_cfg.get('answer_key'),
                default_context_key=data_cfg.get('default_context_key', 'default_context'),
                answer_only_loss=data_cfg.get('answer_only_loss', True),
                is_train=(mode == 'train'),
            )

        logging.warning(
            "!!!You are using legacy dataset and dataloader, please switch to lhotse dataloader "
            "for full functionality and best performance.!!!"
        )
        setattr(self, f"_{mode}_names", data_cfg.get('name', None))

        # Notably, the data weights are controlled by either bucketing_weights
        # or concat_sampling_probabilities depending on the dataset type.
        if data_cfg.get('is_tarred', False):
            dataset = get_tarred_audio_text_dataset_from_config(
                config=data_cfg,
                text_processor=get_text_processor_from_cfg(cfg=self.data_cfg, tokenizer=self.tokenizer),
                augmentor=augmentor,
                global_rank=parallel_state.get_data_parallel_rank(),
                world_size=parallel_state.get_data_parallel_world_size(),
            )
        else:
            dataset = get_audio_text_dataset_from_config(
                manifest_filepath=data_cfg.manifest_filepath,
                config=data_cfg,
                text_processor=get_text_processor_from_cfg(cfg=self.data_cfg, tokenizer=self.tokenizer),
                augmentor=augmentor,
                is_train=(mode == 'train'),
            )
        if mode != 'train':
            num_ds = len(dataset) if isinstance(dataset, list) else 1
            setattr(self, f"_num_{mode}_dl", num_ds)
        return dataset

    def _create_nemo_dataloader(self, dataset: Any, mode: str, **kwargs) -> DataLoader:
        """
        Get dataloader from normal pytorch datasets (non-lhotse)
        """
        self.init_global_step = self.trainer.global_step
        self.data_sampler.init_global_step = self.init_global_step
        data_cfg = self.cfg.get(f"{mode}_ds", None)
        if data_cfg is None:
            logging.info(f"Skipping {mode} dataloader creation as it is not specified in the config: {self.cfg}")
            return None

        if isinstance(dataset, (ConcatMapDataset, BlendableDataset)):
            collate_fn = dataset.datasets[0].collate_fn
        elif hasattr(dataset, 'collate_fn'):
            collate_fn = dataset.collate_fn
        elif hasattr(dataset.datasets[0], 'collate_fn'):
            # support datasets that are lists of entries like ChainDataset
            collate_fn = dataset.datasets[0].collate_fn
        else:
            # support datasets that are lists of lists
            collate_fn = dataset.datasets[0].datasets[0].collate_fn

        if isinstance(dataset, torch.utils.data.IterableDataset):
            data_parallel_size = parallel_state.get_data_parallel_world_size()
            num_micro_batches = data_cfg.global_batch_size // (data_cfg.micro_batch_size * data_parallel_size)
            global_batch_size_on_this_data_parallel_rank = num_micro_batches * data_cfg.micro_batch_size
            # following nemo/collections/llm/gpt/data/fine_tuning.py:FineTuningDataModule:_create_dataloader
            # add WrappedDataLoader to store the dataloader mode
            dataloader = WrappedDataLoader(
                mode=mode,
                dataset=dataset,
                collate_fn=collate_fn,
                shuffle=False,
                batch_size=global_batch_size_on_this_data_parallel_rank,
                drop_last=True,
                num_workers=data_cfg.num_workers,
                pin_memory=data_cfg.pin_memory,
            )
            return dataloader

        return WrappedDataLoader(
            mode=mode,
            dataset=dataset,
            num_workers=data_cfg.num_workers,
            pin_memory=data_cfg.pin_memory,
            persistent_workers=data_cfg.get("persistent_workers", False),
            collate_fn=collate_fn,
            drop_last=mode not in ["test", "predict"],
            shuffle=data_cfg.get("shuffle", False),
            **kwargs,
        )

    def _parse_lhotse_data_name(self, mode: str) -> List[str]:
        """
        Get the dataset names for lhotse datasets
        """
        if mode == 'train':
            return []
        data_cfg = self.cfg.get(f"{mode}_ds", None)
        if data_cfg.get('manifest_filepath', None):
            manifest_filepath = data_cfg.manifest_filepath
            if isinstance(manifest_filepath, str):
                manifest_filepath = [manifest_filepath]
        else:
            input_cfg = data_cfg.input_cfg
            if isinstance(input_cfg, (str, Path)):
                # Resolve /path/to/input_cfg.yaml into config contents if needed.
                input_cfg = OmegaConf.load(input_cfg)
                assert len(input_cfg) == 1, "Only one dataset with multiple manifest paths is supported for eval"
                data_cfg.input_cfg = input_cfg
                # for getting names
                manifest_filepath = [ic.manifest_filepath for ic in input_cfg[0].input_cfg]

        names = []
        for cur_manifest_filepath in manifest_filepath:
            names.append(Path(cur_manifest_filepath).stem)
        logging.info(f"Parsed names for lhotse {mode} dataset: {names}")
        return names

    def _create_lhotse_dataloader(self, dataset: Any, mode: str, **kwargs) -> DataLoader:
        """
        Get lhotse dataloader
        """
        data_cfg = self.cfg.get(f"{mode}_ds", None)
        if data_cfg is None:
            logging.info(f"Skipping {mode} dataloader creation as it is not specified in the config: {self.cfg}")
            return None

        if mode == "train":
            return get_lhotse_dataloader_from_config(
                data_cfg,
                global_rank=parallel_state.get_data_parallel_rank(),
                world_size=parallel_state.get_data_parallel_world_size(),
                dataset=dataset,
            )
        # for eval, we need to create separate dataset so as to report metrics separately
        else:
            dls = []
            if data_cfg.get('manifest_filepath', None):
                manifest_filepath = data_cfg.manifest_filepath
                for cur_manifest_filepath in manifest_filepath:
                    conf = copy.deepcopy(data_cfg)
                    conf['manifest_filepath'] = cur_manifest_filepath
                    dls.append(
                        get_lhotse_dataloader_from_config(
                            conf,
                            global_rank=parallel_state.get_data_parallel_rank(),
                            world_size=parallel_state.get_data_parallel_world_size(),
                            dataset=dataset,
                        )
                    )
            else:
                input_cfg = data_cfg.input_cfg
                if isinstance(input_cfg, (str, Path)):
                    # Resolve /path/to/input_cfg.yaml into config contents if needed.
                    input_cfg = OmegaConf.load(input_cfg)
                    assert len(input_cfg) == 1, "Only one dataset with multiple manifest paths is supported for eval"
                    data_cfg.input_cfg = input_cfg
                for cur_input_cfg in input_cfg[0].input_cfg:
                    conf = copy.deepcopy(data_cfg)
                    conf.input_cfg[0].input_cfg = [cur_input_cfg]
                    dls.append(
                        get_lhotse_dataloader_from_config(
                            conf,
                            global_rank=parallel_state.get_data_parallel_rank(),
                            world_size=parallel_state.get_data_parallel_world_size(),
                            dataset=dataset,
                        )
                    )
            return dls

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """
        Return train dataloader
        """
        data_cfg = self.cfg.get("train_ds", None)
        self.init_global_step = self.trainer.global_step
        self.data_sampler.init_global_step = self.init_global_step
        if data_cfg.get("use_lhotse"):
            return self._create_lhotse_dataloader(self._train_ds, 'train')
        else:
            return self._create_nemo_dataloader(self._train_ds, 'train')

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """
        Return val dataloader
        """
        data_cfg = self.cfg.get("validation_ds", None)
        if data_cfg.get("use_lhotse"):
            data_loaders = self._create_lhotse_dataloader(self._validation_ds, 'validation')
        else:
            if isinstance(self._validation_ds, list):
                if len(self._validation_ds) > 1:
                    data_loaders = [self._create_nemo_dataloader(ds, 'validation') for ds in self._validation_ds]
                else:
                    data_loaders = self._create_nemo_dataloader(self._validation_ds[0], 'validation')
            else:
                data_loaders = self._create_nemo_dataloader(self._validation_ds, 'validation')
        self._num_validation_dl = len(data_loaders) if isinstance(data_loaders, list) else 1
        return data_loaders

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """
        Return test dataloader
        """
        data_cfg = self.cfg.get("test_ds", None)
        if data_cfg.get("use_lhotse"):
            data_loaders = self._create_lhotse_dataloader(self._test_ds, 'test')
        else:
            if isinstance(self._test_ds, list):
                if len(self._test_ds) > 1:
                    data_loaders = [self._create_nemo_dataloader(ds, 'test') for ds in self._test_ds]
                else:
                    data_loaders = self._create_nemo_dataloader(self._test_ds[0], 'test')
            else:
                data_loaders = self._create_nemo_dataloader(self._test_ds, 'test')
        self._num_test_dl = len(data_loaders) if isinstance(data_loaders, list) else 1
        return data_loaders

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        """
        Return predict dataloader
        """
        if "predict_ds" not in self.cfg and "test_ds" in self.cfg:
            data_cfg = self.cfg.get("test_ds", None)
            data_key = 'test'
        elif "predict_ds" not in self.cfg and "validation_ds" in self.cfg:
            data_cfg = self.cfg.get("validation_ds", None)
            data_key = 'validation'
        else:
            data_cfg = self.cfg.get("predict_ds", None)
            data_key = 'predict'

        self._test_ds = self._create_dataset(data_key)
        if data_cfg.get("use_lhotse"):
            return self._create_lhotse_dataloader(self._test_ds, 'predict')
        else:
            if isinstance(self._test_ds, list):
                if len(self._test_ds) > 1:
                    return [self._create_nemo_dataloader(ds, 'predict') for ds in self._test_ds]
                else:
                    return self._create_nemo_dataloader(self._test_ds[0], 'predict')
            else:
                return self._create_nemo_dataloader(self._test_ds, 'predict')

    def state_dict(self) -> Dict[str, Any]:
        """Called when saving a checkpoint, implement to generate and save datamodule state.

        Returns:
            A dictionary containing datamodule state.

        """
        consumed_samples = self.data_sampler.compute_consumed_samples(self.trainer.global_step - self.init_global_step)
        return {'consumed_samples': consumed_samples}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Called when loading a checkpoint, implement to reload datamodule state given datamodule stat

        Args:
            state_dict: the datamodule state returned by ``state_dict``.

        """

        consumed_samples = state_dict['consumed_samples']
        self.data_sampler.init_consumed_samples = consumed_samples
        self.data_sampler.prev_consumed_samples = consumed_samples

        update_num_microbatches(
            consumed_samples=consumed_samples,
            consistency_check=False,
        )
        self.data_sampler.if_first_step = 1
